#include "RdbIngresSqlda.h"


RdbIngresSqlda::RdbIngresSqlda( int items )
{
    sqlda = NULL;
    itemsPrepared = 0;
    Resize( items );
    sqlda->sqld = 0;		// 0 items prepared
}


RdbIngresSqlda::~RdbIngresSqlda()
{
    Unprepare();
    if( sqlda )
      free( sqlda );
}


void RdbIngresSqlda::Resize( int items )
{
    if( itemsPrepared > items ) //cannot resize smaller than prepared size
      items = itemsPrepared;
    sqlda = (IISQLDA*) 
      realloc( sqlda, IISQDA_HEAD_SIZE + (items * IISQDA_VAR_SIZE) );
    sqlda->sqln = items;
}


void RdbIngresSqlda::Prepare()
{
    assert(sqlda);
    int items = sqlda->sqld;
    Unprepare();
    sqlda->sqld = items;
    for( int i = 0 ; i < sqlda->sqld ; i++ ) {
	IISQLVAR* sqv = &sqlda->sqlvar[i];
	int baseType = abs( sqv->sqltype );

	switch( baseType ) {

	    // int types:
	  case IISQ_INT_TYPE:
	    sqv->sqllen = sizeof( RdbInt );
	    break;

	    // float types:
	  case IISQ_MNY_TYPE:
	    if( sqv->sqltype < 0 )
	      sqv->sqltype = -IISQ_FLT_TYPE;
	    else
	      sqv->sqltype = IISQ_FLT_TYPE;
	    /* fall through */
	  case IISQ_FLT_TYPE:
	    sqv->sqllen = sizeof( RdbReal );
	    break;
	    
	    // character types:
	  case IISQ_DTE_TYPE:
	    sqv->sqllen =IISQ_DTE_LEN;
	    /* fall through */
	  case IISQ_CHA_TYPE:
	  case IISQ_VCH_TYPE:
	    sqv->sqllen++; /* for null */
	    if( sqv->sqltype < 0 )
	      sqv->sqltype = -IISQ_CHA_TYPE;
	    else
	      sqv->sqltype = IISQ_CHA_TYPE;
	    break;

	  default:
	    assert(0); // should never get here!
	}
	
	sqv->sqldata = (char*) malloc( sqv->sqllen );

	if( sqv->sqltype < 0 ) { /* need null indicator */
	    sqv->sqlind = (short*) malloc( sizeof(short) );
	    *sqv->sqlind = 0;
	} else {
	    sqv->sqlind = NULL;
	}
    }
    itemsPrepared = sqlda->sqld;
}


void RdbIngresSqlda::Prepare(const RdbIngresSqlda& other,
			     const BitVector& whichCols)
{
    assert( other.Arity() == whichCols.len() );
    int arity = whichCols.count();
    Resize( arity );
    int j = 0;
    for( int i = 0; i < other.Arity() ; i++ ) {
	if( whichCols[i] ) {
	    sqlda->sqlvar[j] = other.sqlda->sqlvar[i];
	    j++;
	}
    }
    sqlda->sqld = arity;
    Prepare();
}


void RdbIngresSqlda::Unprepare()
{
    if( !sqlda )
      return;
    for( int i = 0 ; i < itemsPrepared ; i++ ) {
	IISQLVAR* sqv = &sqlda->sqlvar[i];
	free( sqv->sqldata );
	if( sqv->sqltype < 0 ) { /* free null indicator */
	    free( sqv->sqlind );
	}
    }
    itemsPrepared = 0;
    sqlda->sqld = 0;
}


RdbItem::ItemType RdbIngresSqlda::Type( int i ) const
{
    assert( i < sqlda->sqld );
    IISQLVAR* sqv = &sqlda->sqlvar[i];
    if( sqv->sqlind && *(sqv->sqlind) ) // Nullable & is null
      return RdbItem::RDB_NULL;
    switch( abs( sqv->sqltype ) ) {
      case IISQ_INT_TYPE:
	return RdbItem::RDB_INT;
      case IISQ_FLT_TYPE:
	return RdbItem::RDB_REAL;
      case IISQ_CHA_TYPE:
	return RdbItem::RDB_STRING;
    }
    assert(0);			// should not get here
    return RdbItem::RDB_STRING;	// dummy return
}


void RdbIngresSqlda::SetItem( int i, const RdbItem& val )
{
    assert( sqlda && i < itemsPrepared && i < sqlda->sqld && i >= 0 );
    IISQLVAR* sqv = &sqlda->sqlvar[i];
    if( val.Type() == RdbItem::RDB_NULL ) {
	assert( sqv->sqltype < 0 && sqv->sqlind );	// nullable
	*sqv->sqlind = -1;
    } else {
	if( sqv->sqltype < 0 && sqv->sqlind ) // Nullable
	  *sqv->sqlind = 0;
	switch( abs(sqv->sqltype) ) {
	  case IISQ_INT_TYPE:
	    *(RdbInt*)sqv->sqldata = (RdbInt)val;
	    break;
	  case IISQ_FLT_TYPE:
	    *(RdbReal*)sqv->sqldata = (RdbReal)val;
	    break;
	  case IISQ_CHA_TYPE:
	    {
		String s = (String)val;
		char* valstr = (char*)s;
		int len = strlen(valstr);
		if( len >= sqv->sqllen ) { // make room for bigger string
		    sqv->sqllen = len;
		    sqv->sqldata = (char*)realloc( sqv->sqldata, len+1);
		}
		strcpy( (char*)sqv->sqldata, valstr );
		break;
	    }
	}
    }
}


RdbItem RdbIngresSqlda::operator [](int i) const
{
    assert( sqlda && i < itemsPrepared && i < sqlda->sqld && i >= 0 );
    IISQLVAR* sqv = &sqlda->sqlvar[i];
    switch( Type(i) ) {
      case RdbItem::RDB_INT:
	return RdbItem( *(RdbInt*)sqv->sqldata ); 
      case RdbItem::RDB_REAL:
	return RdbItem( *(RdbReal*)sqv->sqldata ); 
      case RdbItem::RDB_STRING:
	return RdbItem( String((char*)sqv->sqldata) ); 
      case RdbItem::RDB_NULL:
	return RdbItem();
    }
    assert(0);
    return RdbItem();
}

