/*****************************************************************************/
//    VersionedBindEnv.C
/*****************************************************************************/

#include "VersionedBindEnv.h"

/*****************************************************************************/
corTrieNode* corTrieNode::copy() 
{
    corTrieNode *newnode = new corTrieNode;
    *newnode = *this;
    return newnode;
}

/*****************************************************************************/
corTrieNode *cormake_trie(int height) 
{    
    corTrieNode *node = new corTrieNode();
    node->height = height;
    if (height > 1) {
        for (int i=0; i < 2; i++)
        node->pointers[i] = cormake_trie(height-1);
    }
    else {
    node->binding.bindenv = NULL;
    node->binding.expr.init();
    }
    return node;
}

/****************************************************************************/
//
// VersionedBindEnvs:
//
//             Make sure VersionedBindEnv and ArrayBindEnv are used in
//        appropriate places.  It's dangerous to use versioned bindenvs
//        in place of arraybindenvs, and inefficient to do the reverse.
/***************
   NOTE:

   Versioned bindenvs should drop the bindenv name for all local bindings.
   Otherwise there will be problems with versioning.
   Suppose we had a bindenv a, with binding (t,a) for variable X.
   If we create a version a' of a, it should have the binding 
   (t,a') for X, and not (t,a).  Changing the names of the bindenv 
   for all bindings can be very expensive.  Hence we use NULL to refer to
   local bindings, internally.  We fool the external world by adding/deleting
   the bindenv when bindings are lookedup/created.

   Non-local bindings are created during rule application, but are 
   undone later on.  For such bindings we store the bindenv name and 
   do not change it on access.

****************/
//---------------------------------------------------------------------------

#ifdef USE_TRIES

corVersionedBindEnv::corVersionedBindEnv(int size, int )
{
    _limit = size;
    unsigned int hpower = 1;
    int height=1;    // height == 1 ==> leaf node
    while(hpower < size) {
       // hpower <<= 1;
       hpower *= 2;
       height++;
    }
    trie = cormake_trie(height);
    _is_versioned = 1;
#ifdef DEBUG
    corC_ConstructEnvCount++;
    corC_ConstructEnvSize += size;
#endif
}

/*****************************************************************************/
corVersionedBindEnv::corVersionedBindEnv()
{
    _limit = 0;
    trie = NULL;
    _is_versioned = 1;
#ifdef DEBUG
    corC_ConstructEnvCount++;
#endif
}

/*****************************************************************************/
corVersionedBindEnv::~corVersionedBindEnv() {
#ifdef DEBUG
    corC_DestructEnvCount++;
#endif
}

/*****************************************************************************/
#ifdef NOBINDENVHACK
corTerm corVersionedBindEnv::lookup(Variable var) const
#else
corTerm corVersionedBindEnv::lookupV(Variable var) const
#endif
{    
    corTrieNode *node = trie;
    ASSERT(((var < _limit) && (trie != NULL)));
    int varbackup = var;

    for(; node->height > 1; var=varbackup) {
        node = node->pointers[ (var >> (node->height-2)) & 1 ];
    }
    var=varbackup;
    corTerm term = node->binding;

    if (term.bindenv==NULL)
    term.bindenv = (corBindEnv *)this;
    return term;
}

/*****************************************************************************/
#ifdef NOBINDENVHACK
int corVersionedBindEnv::bind(Variable var, corTerm term)
#else
int corVersionedBindEnv::bindV(Variable var, corTerm term)
#endif
{
    corTrieNode** nodep = &trie;
    ASSERT(((var < _limit) && trie != NULL));
    int varbackup = var; // To cover for bug in CFront.

    corTrieNode *newnode;
    while( (*nodep)->height > 1) {
    newnode = (*nodep)->copy();
    (*nodep) = newnode;
    nodep = &(newnode->pointers[ (var >> (newnode->height-2)) & 1 ]);
    var = varbackup;
    }
    (*nodep) = (*nodep)->copy();
    (*nodep)->binding = term;  
    if ( (*nodep)->binding.bindenv == this)
    (*nodep)->binding.bindenv = NULL;
    return 1;
}

/*****************************************************************************/
int corVersionedBindEnv::forcebind(Variable var, corTerm term)
{
    /* Directly bind variable without making a copy of the path.
    *     Useful for initialization of bindenvs
    */
    corTrieNode *node = trie;
    int varbackup = var; // To cover for bug in CFront.

    ASSERT((var < _limit));
    for(; node->height > 1;var=varbackup) {
        node = node->pointers[ (var >> (node->height-2)) & 1 ];
    }
    node->binding = term;
    if (node->binding.bindenv == this)
    node->binding.bindenv = NULL;
    return 1;
}


/*****************************************************************************/
corBindEnv * corVersionedBindEnv::make_version() {
    corVersionedBindEnv *newenv = new corVersionedBindEnv();
    (*newenv) = (*this);
    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
void corVersionedBindEnv::resize(int newsize) {

    if (newsize <= _limit)
    return;
#ifdef DEBUG
    corC_ConstructEnvSize += newsize - _limit;
#endif

    int tmp = 1; // To avoid possible CFront bug, I'm using tmp instead of 1

    if ( newsize <= (tmp << (trie->height-1) )) {
    _limit = newsize;
    return;
    }

    corTrieNode *right, *parent;
    tmp = 1;
    for (; newsize > (tmp << (trie->height-1)) ; tmp = 1) {
     right = cormake_trie(trie->height);
    parent = new corTrieNode();
    parent->pointers[0] = trie;
    parent->pointers[1] = right;
    parent->height = trie->height + 1;
    trie = parent;
    } 
    _limit = newsize;
}

/*****************************************************************************/
int corVersionedBindEnv::all_free() {
   for(register int i=0; i < _limit; i++)
       if (!lookup(i).expr.IsEmpty())
        return 0;
   return 1;
}


/*****************************************************************************/
corBindEnv * corVersionedBindEnv::make_version_and_swap()
{
    corVersionedBindEnv *newenv = new corVersionedBindEnv();
    (*newenv) = (*this);
    //  No need to swap tries, since trie updates are non-destructive

    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
void corVersionedBindEnv::restore(corBindEnv *env)
{
    corVersionedBindEnv *newenv = (corVersionedBindEnv *)env;
    trie = newenv->trie;
    _limit = newenv->_limit;
}

/*****************************************************************************/
corBindEnv* corVersionedBindEnv::copy_shell() {
    // This is a hack -- see note on versioning-hack for more information
    // Copy out the structure, but not the contents
    // Use this very carefully --- deleting bindenvs will delete the
    // bindarray too, so better use ::delete on the original bindenv
    //    rather than delete.
    corVersionedBindEnv *newenv = new corVersionedBindEnv ();
    (*newenv) = (*this);
    return newenv;
}

/*-------------------------------------------------------------------*/
#else     /* Don't USE_TRIES */
/*-------------------------------------------------------------------*/

/*****************************************************************************/
corVersionedBindEnv::corVersionedBindEnv(int size, int incr)
{
    _limit = size;
    bindarray = new corTerm [size];
    _is_versioned = 1;
#ifdef DEBUG
    corC_ConstructEnvCount++;
    corC_ConstructEnvSize += _limit;
#endif
}

/*****************************************************************************/
corVersionedBindEnv::corVersionedBindEnv()
{
    _limit = 0;
    bindarray = NULL;
    _is_versioned = 1;
#ifdef DEBUG
    corC_ConstructEnvCount++;
#endif
}

/*****************************************************************************/
corVersionedBindEnv::~corVersionedBindEnv() {
    delete [] bindarray;
#ifdef DEBUG
    corC_DestructEnvCount++;
    corC_DestructEnvSize += _limit;
#endif
}


/*****************************************************************************/
#ifdef NOBINDENVHACK
corTerm corVersionedBindEnv::lookup(Variable var) const
#else
corTerm corVersionedBindEnv::lookupV(Variable var) const
#endif
{    
    corTerm term = bindarray[var];
    if (term.bindenv==NULL)
    term.bindenv = this;
    return term;
}

/*****************************************************************************/
#ifdef NOBINDENVHACK
int corVersionedBindEnv::bind(Variable var, corTerm term)
#else
int corVersionedBindEnv::bindV(Variable var, corTerm term)
#endif
{
    bindarray[var]=binding;  
    if (bindarray[var].bindenv == this)
    bindarray[var].bindenv = NULL;
    return 1;
}

/*****************************************************************************/
int corVersionedBindEnv::forcebind(Variable var, corTerm binding)
{
    bindarray[var]=binding;  
    if (bindarray[var].bindenv == this)
    bindarray[var].bindenv = NULL;
    return 1;
}

/*****************************************************************************/
corBindEnv * corVersionedBindEnv::make_version() {
    corVersionedBindEnv *newenv = new corVersionedBindEnv(_limit);
    for (register int i=0; i< _limit; i++)
    newenv->bindarray[i] = bindarray[i];
    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
void corVersionedBindEnv::resize(int newsize) {
    corTerm *newbindarray = new corTerm [newsize];
    for(int i=0; i < _limit; i++)
       newbindarray[i] = bindarray[i];
    //bzero(&(newbindarray[_limit]), (newsize-_limit) * sizeof(Term));
    memset(&(newbindarray[_limit]), 0, (newsize-_limit) * sizeof(corTerm));
    corTerm *oldbindarray = bindarray;
    bindarray = newbindarray;
    _limit = newsize;
    delete oldbindarray;
}

/*****************************************************************************/
int corVersionedBindEnv::all_free() {
   for(register int i=0; i < _limit; i++)
       if (!bindarray[i].expr.IsEmpty())
        return 0;
   return 1;
}


/*****************************************************************************/
corBindEnv * corVersionedBindEnv::make_version_and_swap()
{
    // This is a hack --- see note on versioning-hack for more information
    corVersionedBindEnv *newenv = new corVersionedBindEnv(_limit);
    for (register int i=0; i< _limit; i++)
    newenv->bindarray[i] = bindarray[i];

    // Swap bindarrays, so that backpatching trashes the new version.
    //  the old version is still available for reuse.

    corTerm *temparray = bindarray;
    bindarray = newenv->bindarray;
    newenv->bindarray = temparray;
    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
void corVersionedBindEnv::restore(corBindEnv *env)
{
    // This is a hack --- see note on versioning-hack for more information
    // env has to be an VersionedBindEnv
    corVersionedBindEnv *newenv = (corVersionedBindEnv *)env;
    bindarray = newenv->bindarray;
    _limit = newenv->_limit;
}

/*****************************************************************************/
corBindEnv* corVersionedBindEnv::copy_shell() {
    // This is a hack -- see note on versioning-hack for more information
    // Copy out the structure, but not the contents
    // Use this very carefully --- deleting bindenvs will delete the
    // bindarray too, so better use ::delete on the original bindenv
    //    rather than delete.
    corVersionedBindEnv *newenv = new corVersionedBindEnv ();
    newenv->_limit = _limit;
    newenv->bindarray = bindarray;
    return newenv;
}

#endif    /* else case of USE_TRIES */

/*****************************************************************************/

