//**************************************************************************
//    ExecutionEnv.C
//**************************************************************************

#include "ExecutionEnv.h"

#include "Config.h"
#include "CoralError.h"
#include "Globals.h"
#include "BuiltinSymbols.h"

//**************************************************************************
corExecutionEnv::corExecutionEnv(corExecutionEnv *proto,
        FILE *in_file, FILE *out_file, FILE *err_file)
{
    explain_fp = (FILE *) NULL;   // non-NULL only if explain is running

    SM_initialized = proto->SM_initialized;
    C_exec_mode = proto->C_exec_mode;
    C_insert_mode_default = proto->C_insert_mode_default;
    C_quiet_mode_default = proto->C_quiet_mode_default;
    GlobalRelationOptions = proto->GlobalRelationOptions;
    dbg_unify= dbg_subsumes= dbg_indexing= dbg_aggregates= dbg_get_next= 
            proto->dbg_unify;
    dbg_sets = proto->dbg_sets;
    dbg_modules = proto->dbg_modules;
    dbg_deletes = proto->dbg_deletes;
    dbg_ordsearch = proto->dbg_ordsearch;
    dbg_pipelining = proto->dbg_pipelining;
    dbg_backtrack = proto->dbg_backtrack;
    C_hash_buckets_default = proto->C_hash_buckets_default;
    C_delta_buckets_default = proto->C_delta_buckets_default;
    C_incr_ratio_default = proto->C_incr_ratio_default;
    C_max_occupancy_default = proto->C_max_occupancy_default;
    C_max_recursion_depth_default = proto->C_max_recursion_depth_default;
    C_relation_type_default = proto->C_relation_type_default;
    C_connect_exodus_default = proto->C_connect_exodus_default;


    profile_sn = proto->profile_sn;
    profile_scc = proto->profile_scc;
    profile_module = proto->profile_module;


    if (in_file)
        input_file = in_file;
    else
        input_file = proto->input_file;

    if (err_file){
        profile_file = err_file;
        trace_file = err_file;
        error_file = err_file;
    } else {
        profile_file = proto->profile_file;
        trace_file = proto->trace_file;
        error_file = proto->error_file;
    }
    trace_filename = proto->trace_filename;
    error_filename = proto->error_filename;

    if (out_file)
        print_file = out_file;
    else 
        print_file = proto->print_file;
    print_filename = proto->print_filename;
    
    C_answer_style_default = proto->C_answer_style_default;
    C_use_quotes = proto->C_use_quotes;
    C_create_magic_default = proto->C_create_magic_default;
    C_read_coralrc_default = proto->C_read_coralrc_default;
    C_verbose_answer_default = proto->C_verbose_answer_default;
    C_print_prompt_default = proto->C_print_prompt_default;
    C_do_history_default = proto->C_do_history_default;

    C_preprocessing_default = proto->C_preprocessing_default;
    C_rewriting_default = proto->C_rewriting_default;
    C_save_module_default = proto->C_save_module_default;
    C_use_supplementary_magic_default 
            = proto->C_use_supplementary_magic_default;
    C_use_magic_default = proto->C_use_magic_default;
    C_sup_magic_indexing_default = proto->C_sup_magic_indexing_default;
    C_check_subsum_default = proto->C_check_subsum_default;
    C_single_scc_default = proto->C_single_scc_default;
    C_single_answer_default = proto->C_single_answer_default;
    C_convert_functions_default = proto->C_convert_functions_default;
    C_index_deltas_default = proto->C_index_deltas_default;
    C_use_factor_magic_default = proto->C_use_factor_magic_default;
    C_use_return_unify_default = proto->C_use_return_unify_default;
    C_use_pipelining_default = proto->C_use_pipelining_default;
    C_use_ordsearch_default = proto->C_use_ordsearch_default;
    C_use_eager_eval_default = proto->C_use_eager_eval_default;
    C_use_monotonic_default = proto->C_use_monotonic_default;
    C_interactive_mode_default = proto->C_interactive_mode_default;
    C_use_exist_opt_default = proto->C_use_exist_opt_default;
    C_use_factoring_default = proto->C_use_factoring_default;
    C_use_predicate_sn_default = proto->C_use_predicate_sn_default;
    C_non_ground_facts_default = proto->C_non_ground_facts_default;
    C_diskrel_default = proto->C_diskrel_default;
    C_eager_indexing_default = proto->C_eager_indexing_default;
    C_sip_reorder_default = proto->C_sip_reorder_default;
    C_nested_scc_eval_default = proto->C_nested_scc_eval_default;

    stackmark = new corStackMark;
    coral_path = proto->coral_path;

    C_interrupt_raised = proto->C_interrupt_raised;
    quit = proto->quit;
    yield = proto->yield;

    C_at_end_module = proto->C_at_end_module;
    in_query_loop = proto->in_query_loop;
    c_module_separator = proto->c_module_separator;
}

//**************************************************************************
corExecutionEnv::corExecutionEnv()
{
    explain_fp = (FILE *) NULL;   // non-NULL only if explain is running
    init();
}

//**************************************************************************
void corExecutionEnv::init(FILE *in_file, FILE *out_file, FILE *err_file)
{
    SM_initialized = 0;
    C_exec_mode = COR_DOING_INTERPRET;
    C_insert_mode_default = COR_INSERT_MODE_DEFAULT;
    C_quiet_mode_default = 0;
    GlobalRelationOptions = 0;
    dbg_unify= dbg_subsumes= dbg_indexing= dbg_aggregates= dbg_get_next= 0;
    dbg_sets=0;
    dbg_modules = 0;
    dbg_deletes = 0;
    dbg_ordsearch=0;
    dbg_pipelining = 0;
    dbg_backtrack = 0;
    C_hash_buckets_default = COR_DEFAULT_INDEX_BUCKETS;
    C_delta_buckets_default = COR_DEFAULT_DELTA_BUCKETS;
    C_incr_ratio_default = COR_INDEX_INCR_RATIO;
    C_max_occupancy_default = COR_INDEX_OCCUPANCY;
    C_max_recursion_depth_default = COR_MAX_RECURSION_DEPTH;
    C_relation_type_default = COR_RELATION_TYPE_DEFAULT;
    C_connect_exodus_default = COR_CONNECT_EXODUS_DEFAULT;
    
    profile_sn = COR_PROFILE_SN_DEFAULT;
    profile_scc = COR_PROFILE_SN_DEFAULT;
    profile_module = COR_PROFILE_MODULE_DEFAULT;

    if (in_file)
        input_file = in_file;

    if (err_file){
        profile_file = err_file;
        trace_file = err_file;
    }
    trace_filename = "stderr";

    if (out_file)
        print_file = out_file;
    print_filename = "stdout";
    
    error_file = err_file;
    error_filename = "stderr";
    
    C_answer_style_default = COR_ANSWER_STYLE_DEFAULT;
    C_use_quotes = COR_USE_QUOTES_DEFAULT;
    C_create_magic_default = COR_CREATE_MAGIC_DEFAULT;
    C_read_coralrc_default = COR_READ_CORALRC_DEFAULT;
    C_verbose_answer_default = COR_VERBOSE_ANSWER_DEFAULT;
    C_print_prompt_default = COR_PRINT_PROMPT_DEFAULT;
    C_do_history_default = COR_DO_HISTORY_DEFAULT;

    C_preprocessing_default = COR_PREPROCESSING_DEFAULT;
    C_rewriting_default = COR_REWRITING_DEFAULT;
    C_save_module_default = COR_SAVE_MODULE_DEFAULT;
    C_use_supplementary_magic_default = COR_USE_SUPPLEMENTARY_MAGIC_DEFAULT ;
    C_use_magic_default = COR_USE_MAGIC_DEFAULT;
    C_sup_magic_indexing_default = COR_USE_SUP_MAGIC_INDEXING_DEFAULT;
    C_check_subsum_default = COR_CHECK_SUBSUM_DEFAULT;
    C_single_scc_default = COR_SINGLE_SCC_DEFAULT;
    C_single_answer_default = COR_SINGLE_ANSWER_DEFAULT;
    C_convert_functions_default = COR_CONVERT_FUNCTIONS_DEFAULT;
    C_index_deltas_default = COR_INDEX_DELTA_DEFAULT;
    C_use_factor_magic_default = COR_USE_FACTOR_MAGIC_DEFAULT;
    C_use_return_unify_default = COR_USE_RETURN_UNIFY_DEFAULT;
    C_use_pipelining_default = COR_USE_PIPELINING_DEFAULT;
    C_use_ordsearch_default = COR_USE_ORDSEARCH_DEFAULT;
    C_use_eager_eval_default = COR_USE_EAGER_EVAL_DEFAULT;
    C_use_monotonic_default = COR_USE_MONOTONIC_DEFAULT;
    C_interactive_mode_default = COR_INTERACTIVE_MODE_DEFAULT;
    C_use_exist_opt_default = COR_USE_EXIST_OPT_DEFAULT;
    C_use_factoring_default = COR_USE_FACTORING_DEFAULT;
    C_use_predicate_sn_default = COR_USE_PREDICATE_SEMINAIVE_DEFAULT;
    C_non_ground_facts_default = COR_NON_GROUND_FACTS_DEFAULT;
    C_diskrel_default = COR_DISKREL_DEFAULT;
    C_eager_indexing_default = COR_EAGER_INDEXING_DEFAULT;
    C_sip_reorder_default = COR_SIP_REORDER_DEFAULT;
    C_nested_scc_eval_default = COR_NESTED_SCC_EVAL_DEFAULT;  // added by Ken
    
    stackmark = new corStackMark ;
    coral_path = getenv("CORALROOT");
    if (!coral_path) {
#        ifdef CORALROOT
            coral_path = CORALROOT;
#        else
            fprintf(error_file,
                " Warning! CORALROOT not set in environment!\n");
            coral_path = DEFAULT_CORAL_PATH;
            fprintf(error_file, " Using %s\n", DEFAULT_CORAL_PATH);
#        endif
    }

    C_interrupt_raised = 0;
    quit = 0;
    yield = 0;

    C_at_end_module = 1;
    in_query_loop = 1;
    c_module_separator = 0;
}

//**************************************************************************
corExecutionEnv::~corExecutionEnv()
{
    delete stackmark;
}
//**************************************************************************
void corExecutionEnv::flush_all_outputs()
{
      // Do not close files in server mode, the connection handles it.
    if (! coralConfig.server_mode){

        if ((trace_file) && (trace_file != stdout)
            && (trace_file != stderr))
          fclose(trace_file);
        
        if ((print_file) && (print_file != stdout)
            && (print_file != stderr))
          fclose(print_file);
        
        if ((error_file) && (error_file != stdout)
            && (error_file != stderr))
          fclose(error_file);

#ifdef    DO_EXPLAIN
        if (explain_fp)
            fclose(explain_fp);
#endif    

    }
}

//**************************************************************************
void corExecutionEnv::display(FILE *outf)
{

    fprintf(outf, "\tmode :\t\t\t");

    switch (C_exec_mode) {
    case COR_DOING_INTERPRET:
        fprintf(outf, "INTERPRETER\n");
        break;
    case COR_DOING_COMPILE:
        fprintf(outf, "COMPILER\n");
        break;
    case COR_DOING_MAGIC:
        fprintf(outf, "MAGIC Rewriting\n");
        break;
    case COR_DOING_FACTOR:
        fprintf(outf, "FACTOR Rewriting\n");
        break;
    case COR_DOING_EXIST:
        fprintf(outf, "Existential Query Optimization\n");
        break;
    case COR_DOING_IMPERATIVE:
        fprintf(outf, "Imperative module\n");
        break;
    case COR_DOING_FACTOR_MAGIC:
        default:
        fputs("This can't happen\n",stderr);
        break;
    }

    fprintf(outf, "\tinsert_mode : \t\t%s\n",
        (C_insert_mode_default) ? "on" : "off");
    fprintf(outf, "\tquiet_mode : \t\t%s\n",
        (C_quiet_mode_default) ? "on" : "off");

    fprintf(outf,"--- Defaults for Rewriting ---\n\n");

    fprintf(outf, "\trewriting : \t\t%s\n",
        (C_rewriting_default) ? "on" : "off");
    fprintf(outf, "\tsup_magic :\t\t%s\n",
        (C_use_supplementary_magic_default) ? "on" : "off");
    fprintf(outf, "\tmagic :\t\t\t%s\n",
        (C_use_magic_default) ? "on" : "off");
    fprintf(outf, "\tsup_magic_indexing :\t%s\n",
        (C_sup_magic_indexing_default) ? "on" : "off");
    fprintf(outf, "\texistential : \t\t%s\n",
        (C_use_exist_opt_default) ? "on" : "off");
    fprintf(outf, "\tfactoring : \t\t%s\n\n",
        (C_use_factoring_default) ? "on" : "off");

    fprintf(outf,"--- Defaults for Execution ---\n\n");

    fprintf(outf, "\tinteractive_mode :\t%s\n",
        (C_interactive_mode_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t\t%s\n", corSymbolString(corPredicateSNSymbol),
        (C_use_predicate_sn_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corUsePipeliningSymbol),
        (C_use_pipelining_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corUseOrdSearchSymbol),
        (C_use_ordsearch_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corEagerEvalSymbol),
        (C_use_eager_eval_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corSaveModuleSymbol),
        (C_save_module_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corMonotonicSymbol),
        (C_use_monotonic_default) ? "on" : "off");

    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corMultisetSymbol),
        (C_multisets_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corCheckSubsumSymbol),
        (C_check_subsum_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corIndexDeltaSymbol),
        (C_index_deltas_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corReturnUnifySymbol),
        (C_use_return_unify_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corNonGroundFactsSymbol),
        (C_non_ground_facts_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corDiskResidentSymbol),
        (C_diskrel_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corEagerIndexingSymbol),
        (C_eager_indexing_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corSipReorderSymbol),
        (C_sip_reorder_default) ? "on" : "off");
    fprintf(outf, "\t%s :\t%s\n", corSymbolString(corNestedSccEvalSymbol),
        (C_nested_scc_eval_default) ? "on" : "off"); // added by Ken

    fprintf(outf, "\n--- Defaults for Output ---\n");
    fprintf(outf, "\tprint_file :\t\t%s\n", print_filename);
    fprintf(outf, "\ttrace_file :\t\t%s\n", trace_filename);
    fprintf(outf, "\ttracing :\t\t%s\n",
        (corexEnv->GlobalRelationOptions & REL_DISPLAY_INSERTIONS) ? "on" : "off");
    fprintf(outf, "\tprofiling :\t\t%s\n",
        (corexEnv->GlobalRelationOptions & REL_PROFILE_STATS) ? "on" : "off");
    fprintf(outf, "\texplain :\t\t%s\n",
        (corexEnv->GlobalRelationOptions & REL_EXPLAIN) ? "on" : "off");
    fprintf(outf, "\tanswer_style :\t%s\n",
        (C_answer_style_default) ? "binding" : "tuple");
    fprintf(outf, "\tuse quotes :\t%s\n", (C_use_quotes) ? "yes" : "no");
    fprintf(outf, "\tverbose_answer :\t%s\n",
        (C_verbose_answer_default) ? "on" : "off");
    fprintf(outf, "\tprint_prompt :\t%s\n",
        (C_print_prompt_default) ? "on" : "off");
    fprintf(outf, "\tcreate_magic :\t%s\n",
        (C_create_magic_default) ? "on" : "off");
    fprintf(outf, "\thistory :\t%s\n",
        (C_do_history_default) ? "on" : "off");

    fprintf(outf, "\n--- Defaults for Profiling ---\n");
    fprintf(outf, "\tprofile_module :\t%s\n",
        (profile_module) ? "on" : "off");
    fprintf(outf, "\tprofile_scc :\t\t%s\n",
        (profile_scc) ? "on" : "off");
    fprintf(outf, "\tprofile_sn :\t\t%s\n",
        (profile_sn) ? "on" : "off");

#    ifdef DEBUG
        fprintf(outf, "\n--- Defaults at System Level ---\n");

        fprintf(outf, "\tconnect_exodus: \t%s\n",
            (C_connect_exodus_default) ? "on" : "off");
        fprintf(outf, "\tpreprocessing : \t%s\n",
            (C_preprocessing_default) ? "on" : "off");
        fprintf(outf, "\t%s :\t\t%s\n", corSymbolString(corSingleSccSymbol),
            (C_single_scc_default) ? "on" : "off");
        fprintf(outf, "\t%s :\t%s\n", corSymbolString(corConvertFunctionsSymbol),
            (C_convert_functions_default) ? "on" : "off");

        fprintf(outf, "\n--- Defaults for System Debugging ---\n");
        fprintf(outf, "\tdebug_unify :\t\t%s\n",
            (dbg_unify) ? "on" : "off");
        fprintf(outf, "\tdebug_subsumes :\t%s\n",
            (dbg_subsumes) ? "on" : "off");
        fprintf(outf, "\tdebug_indexing :\t%s\n",
            (dbg_indexing) ? "on" : "off");
        fprintf(outf, "\tdebug_aggregates :\t%s\n",
            (dbg_aggregates) ? "on" : "off");
        fprintf(outf, "\tdebug_sets :\t\t%s\n",
            (dbg_sets) ? "on" : "off");
        fprintf(outf, "\tdebug_get_next :\t%s\n",
            (dbg_get_next) ? "on" : "off");
        fprintf(outf, "\tdebug_ordered_search :\t%s\n",
            (dbg_ordsearch) ? "on" : "off");
        fprintf(outf, "\tdebug_pipelining :\t%s\n",
            (dbg_pipelining) ? "on" : "off");
        fprintf(outf, "\tdebug_backtrack :\t%s\n",
            (dbg_backtrack) ? "on" : "off");


        fprintf(outf, "\n--- Defaults for System Tuning ---\n");
        fprintf(outf, "\thash_buckets :\t\t%d\n",C_hash_buckets_default);
        fprintf(outf, "\tdelta_buckets :\t\t%d\n", C_delta_buckets_default);
        fprintf(outf, "\tincr_ratio :\t\t%f\n",C_incr_ratio_default);
        fprintf(outf, "\tmax_occupancy :\t\t%f\n",C_max_occupancy_default);
        fprintf(outf, "\trecursion_depth :\t%d\n",
            C_max_recursion_depth_default);
        fprintf(outf, "\trelation_type :\t\t%d\n", C_relation_type_default);
#    endif 
}

//**************************************************************************
