/*****************************************************************************/
//    HashIndexSet.C
/*****************************************************************************/

#include "HashIndexSet.h"

#include "Interface.h"
#include "Globals.h"
#include "HashSimpleRelation.h"

/*****************************************************************************/
/*
    WARNING::
    Problem in HashIndexSet::find_exact
    re-orderings of var_names is not taken into account while checking
    whether two indexes are the same.  Thus, [V0|V1],V2 ; V1,V2 is
    treated differently from [V0|V1],V2 ; V2,V1 though the two should
    really be the same.  Merely sorting the var_names will not do the
    trick since [V0|V1],V2 ; V1,V2 subsumes and is subsumed by
    [V2|V1],V0 ; V1,V0 but sorting var_names would mess that up.
*/

int corHashIndexSet::find_exact( corHashIndex *index )
{
    for (int i = 0; i < num_in_array; i++) {
        if (index->argset != NULL) {
            // argument form indexing
            if ( ((HashIndexPtr) Indices[i])->argset != NULL &&
                *index->argset == *((HashIndexPtr) Indices[i])->argset)
            {
                return COR_I_SUCCESS;
            }
        }
        else {
            // pattern form indexing
            if ( ((HashIndexPtr) Indices[i])->argset == NULL &&
                (index->num_vars_pat ==
                    ((HashIndexPtr)Indices[i])->num_vars_pat) &&
                (index->var_names->count() ==
                    ((HashIndexPtr) Indices[i])->var_names->count()))
            {
                corStackMark stackmark;
                
                corArrayBindEnv pat_env1(index->num_vars_pat);
                corArrayBindEnv pat_env2(index->num_vars_pat);
                corArrayBindEnv pat_env3(index->num_vars_pat);
                corArrayBindEnv pat_env4(index->num_vars_pat);
                        
                //    check if pattern and the var_names of each subsume
                //    that of the other
                
                if (
                    (corsubsumes_arg_lists(
                        *concat_arglists(index->pattern,index->var_names),
                        *concat_arglists(((HashIndexPtr) Indices[i])->pattern,
                        ((HashIndexPtr) Indices[i])->var_names),
                        &pat_env1,
                        &pat_env2) == COR_U_SUCCEED)
                    &&
                    (corsubsumes_arg_lists(
                        *concat_arglists(((HashIndexPtr) Indices[i])->pattern,
                            ((HashIndexPtr) Indices[i])->var_names),
                        *concat_arglists(index->pattern, index->var_names),
                        &pat_env3,
                        &pat_env4) == COR_U_SUCCEED)
                    )
                {
                    stackmark.pop_to();
                    return COR_I_SUCCESS;
                }
                else {
                    stackmark.pop_to();
                }
            }
        }
    }
    return COR_I_FAIL;
}

/*****************************************************************************/
//    This is an optimization to let a static environment be used in 
//    reentrant code.  We assume no parallelism however.

static corArrayBindEnv static_env(MAX_VARS,VAR_INCR);
static int static_env_inuse=DONT_USE_STATIC_ENV;

/*****************************************************************************/
/*
    find an index that best matches the pattern contained arg_list, bind_env

    WARNING:: The following code should at least check if the index is 
    appropriate.    What it does is rather dangerous.  This is inappropriate
    for programs with non-ground terms.  What is an appropriate index for one
    invocation may not be appropriate for subsequent invocations : PRAVEEN
*/

corGenericIndex *corHashIndexSet::find_index(corArgList &arg_list, corBindEnv *bindenv,
    int *tryindexp)
{
    if ((tryindexp) && ((*tryindexp) >= 0) && ((*tryindexp) < num_in_array)) {
        // PRAVEEN: i checked that this gets entered, and yes it does !
        return ((corGenericIndex *) Indices[*tryindexp]);
    }    

    if (!num_in_array) {
        fprintf(corexEnv->error_file, "There are no indices: Quitting\n");
        return NULL;
    }

    // find bitvector of pattern 
    int arity = arg_list.count();

    if (arity == 0) {
        // propositional case.
        if (rel->arity() != 0) {
            fprintf(corexEnv->error_file,
                "Error: HashIndexSet::find_index: arity mismatch\n");
        }
        //    Arbitrarily choose the 1st index and do a sequential scan.
        //    need to ensure that the 1st index is an argument-type index.
        if (tryindexp) *tryindexp = 0;
        return (corGenericIndex *) Indices[0];
    }

    corBitVector adornment(arity);
    int num_bound = 0; // number of bound arg positions in pattern
    for (int i = 0; i < arity; i++) {
        int bound = 1;
        // check which arguments are ground and set those bits to 1; rest 0
        if (arg_list[i].hash(bindenv) == VarHashValue)
            bound = 0;
        if (bound == 1) num_bound++;
        adornment.set(i, bound);
    }

    //    Need to find "best" match 
    //    For argument form indices: only those with FEWER ground arguments
    //        considered: i.e. index pattern subsumes pattern
    //    For pattern form indices: index pattern should subsume pattern.
    //    If all hash variables bound: give appropriate high weight
    //    If only some bound: give a small weight say 0.5 because of
    //        greater selectivity than other trivial indices.

    corGenericIndex *return_index = NULL;
    double max_weight = -1.0;

    for (i = 0; i < num_in_array; i++) {
        double cur_weight = 0.0;
        if (((HashIndexPtr) Indices[i])->argset != NULL) {
            // argument index
            if (arity != ((HashIndexPtr) Indices[i])->argset->len())
                continue;

            if ((num_bound == 0) &&
                !(((HashIndexPtr)Indices[i])->argset->any()))
            {
                // all free index
                if (tryindexp) *tryindexp = i;
                return ((corGenericIndex *)Indices[i]);                
            }

            if ((num_bound == arity) &&
                (adornment == *(((HashIndexPtr) Indices[i])->argset)))
            {
                // All arguments are bound - no better match can be found.
                if (tryindexp) *tryindexp = i;
                return ((corGenericIndex *)Indices[i]) ;
            }
            for (int j = 0; j < arity; j++) {
                if (adornment.test(j) &&
                    ((HashIndexPtr) Indices[i])->argset->test(j))
                {
                    cur_weight += 1.0;    // both are bound
                    continue;
                }
                if (((HashIndexPtr) Indices[i])->argset->test(j)) {
                    // and !adornment.test(j)
                    // the index has more bound arguments; it is useless
                    cur_weight = 0.0; 
                    break;
                }
            }
        }
        else {
            // it is a pattern index
            corStackMark stack_mark;
            corArrayBindEnv *patenv1;

#            define CLEANUP { \
                stack_mark.pop_to(); \
                if (--static_env_inuse != 0 || \
                    ((HashIndexPtr) Indices[i])->num_vars_pat > MAX_VARS) \
                { \
                    delete patenv1; \
                } \
            }

            if (static_env_inuse++ == 0 &&
                ((HashIndexPtr) Indices[i])->num_vars_pat <= MAX_VARS)
            {
                patenv1 = &static_env;
            }
            else {
                patenv1 = new corArrayBindEnv 
                    (((HashIndexPtr) Indices[i])->num_vars_pat);
            }

            if (corsubsumes_arg_lists(*((HashIndexPtr) Indices[i])->pattern,
                arg_list, patenv1, bindenv) == COR_U_SUCCEED)
            {
                for (int j = 0;
                    j < ((HashIndexPtr) Indices[i])->var_names->count(); 
                     j++)
                {
                    //    The following is a heuristic for deciding what weight
                    //    to assign to pattern form matches.  This will need to
                    //    be updated as and when need be.

                  if ((*((HashIndexPtr) Indices[i])->var_names)[j].hash(patenv1)
                                == VarHashValue)
                    {
                        cur_weight = 0.5;
                        CLEANUP
                        break;
                    }
                    cur_weight += 1.0;
                    if (cur_weight > arity) {
                        cur_weight = arity;
                        CLEANUP
                        break;
                    }
                }
            }
            else {
                //    The pattern form index does not subsume pattern
                cur_weight = 0.0;
                CLEANUP
            }
        }
        if (cur_weight > max_weight) {
            //    need to update information
            max_weight = cur_weight;
            if (tryindexp) *tryindexp = i;
            return_index = (corGenericIndex *) Indices[i];
        }
    }
        
    if (!return_index) {
        //    Have not been able to find any match
        //    Arbitrarily choose the 1st index and do a sequential scan.
        //    need to ensure that the 1st index is an argument-type index.
        if (tryindexp) *tryindexp = 0;
        return_index = (corGenericIndex *) Indices[0];
    }

    return return_index;
}

/*****************************************************************************/
int corHashIndexSet::insert_index( corGenericIndex *index )
{

    corHashIndex *h_index = (corHashIndex *) index;
    if (find_exact(h_index )) {
#        if DO_TRACE
            if (corexEnv->GlobalRelationOptions & REL_DISPLAY_INDEXOPS) {
                fprintf(corexEnv->trace_file, "Duplicate index: not inserted\n");
            }
#        endif
        return(COR_I_FAIL);
    }
    else {
        // new index to be inserted in list
        if (num_in_array == max_in_array) {
#            ifdef DEBUG
                fprintf(corexEnv->error_file, "Too many indices: not inserted\n");
#            endif
              return COR_I_FAIL;
        }
        Indices[num_in_array++] = (corGenericIndex *) h_index;
        // this is fine since array insertion starts from zero.
        return COR_I_SUCCESS;
    }
}

/*****************************************************************************/
/*
    index is the index that will be used to fill up h_index. 
    it should be an argument form index or a pattern form
    index that indexes ALL the tuples in the relation, else
    the filling up will not be correct.
*/

void corHashIndexSet::fill_index( corGenericIndex *index )
{
    corHashIndex *h_index = (corHashIndex *) index;
    corHashIndex *f_index = (corHashIndex *) Indices[0];

    //    Don't want to fill up an index using itself
    ASSERT(h_index != f_index);

    //    Now sequentially scan through f_index
    corTuple *tuple = NULL;
    corHashNode *cur_pos = NULL;
    for (int i = 0; i < f_index->hash_size; i++) {
        for (cur_pos = f_index->hash_table[i]; cur_pos != NULL; 
            cur_pos=cur_pos->next)
        {
            h_index->insert_tuple(cur_pos->tuple());
        }
    }
    for (cur_pos = f_index->var_chain; cur_pos != NULL;
        cur_pos = cur_pos->next)
    {
        h_index->insert_tuple(cur_pos->tuple());
    }
}

/*****************************************************************************/
int corHashIndexSet::add_all_tuples(corStorageRelation *_reln,
    int marks_into_account, int check_dupl, int hide_em)
{
    //    WARNING: duplicate checking is done only when marks are taken
    //    into account.

    corHashIndex *f_index = (corHashIndex *) Indices[0];
    int dupl_check = check_dupl && _reln->check_subsum;

    if (!f_index->is_argument_form()) {
        fprintf(corexEnv->error_file,
            "ERROR: add_all_tuples does not have proper index.\n");
        return COR_I_FAIL;
    }

    corTuple *tuple = NULL;
    corHashNode *cur_pos = NULL;

    register int j;

    if (!marks_into_account) {
        // Move nodes in the hash buckets

        for (j=f_index->non_empty_bucket_head; j >= 0;
            j=f_index->non_empty_buckets[j])
        {
            for (cur_pos = f_index->hash_table[j]; cur_pos; 
                cur_pos = cur_pos->next)
            {
                  if (!hide_em)
                    _reln->insert_raw(cur_pos->tuple());
                  else
                    _reln->insert_in_hidden_relation(cur_pos->tuple());
            }
        }

        for (cur_pos = f_index->var_chain; cur_pos != NULL; 
            cur_pos = cur_pos->next)
        {
            if (!hide_em)
                _reln->insert_raw(cur_pos->tuple());
            else
                _reln->insert_in_hidden_relation(cur_pos->tuple());
        }
    }
    else {
        // same as above code except that insert_tuple has to be used

        for (j=f_index->non_empty_bucket_head; j >= 0 ;
            j=f_index->non_empty_buckets[j])
        {
            for (cur_pos = f_index->hash_table[j];cur_pos;
                cur_pos=cur_pos->next)
            {
                if ((!dupl_check) || (!_reln->is_subsumed(cur_pos->tuple()))) {
                    if (!hide_em)
                        _reln->insert_tuple(cur_pos->tuple());
                    else
                        _reln->insert_in_hidden_relation(cur_pos->tuple());
                }
            }
        }
      
        for (cur_pos = f_index->var_chain; cur_pos != NULL; 
            cur_pos = cur_pos->next)
        {
            if ((!dupl_check) || (! _reln->is_subsumed(cur_pos->tuple()))) {
                if (!hide_em)
                    _reln->insert_tuple(cur_pos->tuple());
                else
                    _reln->insert_in_hidden_relation(cur_pos->tuple());
            }
        }
    }

    return COR_I_SUCCESS;
}

/*****************************************************************************/
int corHashIndexSet::add_all_goals(corStorageRelation *_reln, corContext *_context,
    corStorageRelation *owner_rel)
{
    //    This is used in Ordered Search
    //    Tuples in (this) need to be converted to GoalNodes and added to
    //        reln as well as context

    corHashIndex *f_index = (corHashIndex *) Indices[0];

    if (!f_index->is_argument_form()) {
        fprintf(corexEnv->error_file,
            "ERROR: add_all_goals does not have proper index.\n");
        return COR_I_FAIL;
    }

    corTuple *tuple = NULL;
    corGoalNode *g_node = NULL;
    corHashNode *cur_pos = NULL;
    for (int i = 0; i < f_index->hash_size; i++) {
        for (cur_pos = f_index->hash_table[i]; cur_pos != NULL;
            cur_pos=cur_pos->next)
        {
            tuple = cur_pos->tuple();
            if (!(tuple->is_deleted())) {
                g_node = new corGoalNode(tuple->_args, tuple->bindenv, owner_rel);
                g_node->env_size = tuple->env_size;
                _reln->insert_tuple(g_node);
                _context->add_new_goal(g_node, NULL);
            }
        }
    }

    for (cur_pos = f_index->var_chain; cur_pos != NULL; 
        cur_pos = cur_pos->next)
    {
        tuple = cur_pos->tuple();
        if (!(tuple->is_deleted())) {
            g_node = new corGoalNode(tuple->_args, tuple->bindenv, owner_rel);
            g_node->env_size = tuple->env_size;
            _reln->insert_tuple(g_node);
            _context->add_new_goal(g_node, NULL);
        }
    }
    return COR_I_SUCCESS;
}

/*****************************************************************************/
int corHashIndexSet::move_all_entries(corStorageRelation *_reln)
{
    /*
        WARNING: This code assumes that the two relations have exactly the
        same indices in exactly the same order, and both the relations
        are HashSimpleRelations
    */

    corHashSimpleRelation *otherrel = (corHashSimpleRelation *) _reln;
    corHashIndexSet *otherindset =  &(otherrel->index_set);
    
    if(num_in_array != otherindset->num_in_array) {
        fprintf(corexEnv->error_file,
            "CORAL:: internal error in move_all_entries\n");
        return -1;
    }

    for (int i=0; i < num_in_array; i++) {
        Indices[i]->move_all_entries(otherindset->Indices[i]);
    }
    return 1;
}

/*****************************************************************************/

