/*****************************************************************************/
//    PipelinedDerivedRelation.C
/*****************************************************************************/

#include "PipelinedDerivedRelation.h"

#include "QFPipelinedModuleData.h"
#include "Compile.h"
#include "TupleIterator.h"
#include "Globals.h"
#include "Execute.h"
#include "TableStack.h"

/*****************************************************************************/

static void initRulePointers(corTable* F, corQFModuleData* md);

/*****************************************************************************/
corRelation* cormakePipelinedRelation(corQFPipelinedModuleData *md,Name name, int arity)
{
    corRelation *rel = new corPipelinedDerivedRelation(md,name,arity);
    return rel;
}

/*****************************************************************************/
corPipelinedDerivedRelation::corPipelinedDerivedRelation(
    corQFPipelinedModuleData *module,Name _name,int new_arity)
{
    md = module;
    name = _name;
    _arity = new_arity;
    methods = NULL;
    r_kind = COR_R_DERIVED;
    has_release = 1;
    pred_offset = corget_offset(module->F, name, _arity);
}

/*****************************************************************************/
void corPipelinedDerivedRelation::release(corTupleIterator& iterator)
{
    // get rid of the state of the iterator
    if (iterator.pipe) {
        delete iterator.pipe;
        iterator.pipe = NULL;
    }
}

/*****************************************************************************/
corBindEnv* corPipelinedDerivedRelation::get_next(corTupleIterator& iterator)
{
    this->get_next_tuple(iterator);
    if (iterator.no_match()) return NULL;
    return(iterator.bindenv);
}

/*****************************************************************************/
#define FAIL_RETURN \
{ \
    if (corCurrModuleInfo != iterator.prev_module_info) { \
        corCurrModuleInfo = iterator.prev_module_info; \
        corT_Stack.pop(); \
    } \
    iterator.prev_module_info = NULL; \
    return NULL; \
} 

corTuple* corPipelinedDerivedRelation::get_next_tuple(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    iterator.tuple_env = NULL;
    corArgList *tempArg;
    corArgList newArg;

#    ifdef DO_PROFILE
        // profiling info
        (corGlobalProfileInfo.getnexts)++;
#    endif

    // Check if this is the first time that this iterator has been called
    // If so then perform initialization
    if (!iterator.pipe) {
#        ifdef DO_TRACE
             if (corexEnv->dbg_get_next) {
               fprintf(corexEnv->trace_file,
                   "PipelinedDerivedRelation: new get_next called on %s(", 
                   name->string());
               iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
               fprintf(corexEnv->trace_file, ")\n");
              }
#        endif

        if (corexEnv->C_interrupt_raised) {
            iterator.set_no_match();
            return NULL;
        }

        // if this is the first exec on the module, 
        // we have to set up the Table
        //if (!md->TableSetUp)

        if (!md->TableSetUp++) initRulePointers(md->F,md);

        /*
            Since the Table is being shared by all invocations on this
            module, we can change it directly instead of having to 
            make copies of it for each execution
            ::: this comment is invalid :: PRAVEEN
            Since we would like to maintain an actual stack that
            reflects the module invocations, we need to make copies
            of the Table.
        */

        // This is the first time get_next is being called on this iterator
        // Time to set up PipelinedExecInfo

#        ifdef DO_TRACE
            if (corexEnv->dbg_get_next) {
                fprintf(corexEnv->trace_file,
                "PipelinedDerivedRelation: starting on %s(",
                name->string());
                iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
                fprintf(corexEnv->trace_file, ")\n");
            }
#        endif

        iterator.pipe = new corPipelinedExecInfo(md, pred_offset);
        if (!iterator.pipe->curRule) {
            // if there is no rule defining the predicate
            iterator.set_no_match();
            delete iterator.pipe;
            iterator.pipe = NULL;
#            ifdef DO_TRACE    
                if (corexEnv->dbg_get_next) {
                    fprintf(corexEnv->trace_file,
                        "PipelinedDerivedRelation: new get_next failed on %s(",
                        name->string());
                    iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
                    fprintf(corexEnv->trace_file, ")\n");
                }
#            endif
            return NULL;
        }

        // unify the arg_lists of the iterator and the head of the rule
        // keep trying all the rules till either unification succeeds or 
        // we have exhausted all the rules that define this relation
    
        // get a mark so that we can undo the changes if it fails
        iterator.pipe->stackmark0.get_mark();
        // make an ArgList* from Arg***
        // that points to the ArgList for the rule head
    
        tempArg = (iterator.pipe)->rInfo->arg_list;
    
        // sanity check
        ASSERT(iterator.pipe->rInfo->num_literals ==
            iterator.pipe->curRule->num_preds);
        ASSERT(env != NULL);

        newArg = (tempArg[(iterator.pipe)->rInfo->num_literals]);
        while (corunify_arg_lists( newArg, iterator.arg_list,iterator.pipe->env,
            iterator.bindenv) != COR_U_SUCCEED)
        {
            // unification failed - backtrack
            iterator.pipe->stackmark0.pop_to();
            // Try the next rule
            if (!(iterator.pipe->NextRule(md))) {
                // if we have exhausted all the rules, give up
                iterator.set_no_match();
                delete iterator.pipe;
                iterator.pipe = NULL;
#                ifdef DO_TRACE    
                    if (corexEnv->dbg_get_next) {
                        fprintf(corexEnv->trace_file,
                         "PipelinedDerivedRelation: new get_next failed on %s(",
                         name->string());
                        iterator.arg_list.print(iterator.bindenv,
                            corexEnv->trace_file);
                        fprintf(corexEnv->trace_file, ")\n");
                    }
#                endif
                return NULL;
                //FAIL_RETURN;
                   }
            tempArg = (iterator.pipe)->rInfo->arg_list;
            newArg = tempArg[(iterator.pipe)->rInfo->num_literals];
               }
    }
    else {
        // iterator.pipe->stackmark0.pop_to();
#        ifdef DO_TRACE
            if (corexEnv->dbg_get_next) {
                fprintf(corexEnv->trace_file,
                    "PipelinedDerivedRelation: get_next repeat call on %s(", 
                    name->string());
                iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
                fprintf(corexEnv->trace_file, ")\n");
            }
#        endif
    }

    // check to see that an interrupt has not been raised
    if (corexEnv->C_interrupt_raised) {
#        ifdef DO_TRACE
            if (corexEnv->dbg_get_next) {
                fprintf(corexEnv->trace_file,
                "PipelinedDerivedRelation: get_next failed on %s(",
                name->string());
                iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
                fprintf(corexEnv->trace_file,") due to interrupt\n");
            }
#        endif
        iterator.set_no_match();
        delete iterator.pipe;
        iterator.pipe = NULL;
        return NULL;
        //FAIL_RETURN;
    }
    
    iterator.prev_module_info = corCurrModuleInfo;
    if (corCurrModuleInfo != &(md->module_info)) {
        corT_Stack.push(md->F);
        corCurrModuleInfo = &(md->module_info);
    }
    
    // Now go ahead and call Solve on PipelinedExecInfo
    // keep trying as long as no match is found
    while (!((iterator.pipe)->Solve())) {
        iterator.pipe->stackmark0.pop_to();
        // try the next rule 
        if ( ((iterator.pipe)->cut_crossed) || !(iterator.pipe->NextRule(md)))
        {
            // if we have exhausted all the rules,
            // or have failed because one rule tried to backtrack
            // past a cut, give up
            iterator.set_no_match();
            delete iterator.pipe;
            iterator.pipe = NULL;

#            ifdef DO_TRACE
                if (corexEnv->dbg_get_next) {
                    fprintf(corexEnv->trace_file,
                        "PipelinedDerivedRelation: get_next failed on %s(",
                        name->string());
                    iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
                    fprintf(corexEnv->trace_file,")\n");
                }
#            endif
            //return NULL;
            FAIL_RETURN;
        }
        tempArg = (iterator.pipe)->rInfo->arg_list;
        newArg = (tempArg[(iterator.pipe)->rInfo->num_literals]);

        // See if the head of this rule unifies with the iterator
        while (corunify_arg_lists( newArg, iterator.arg_list,
            (iterator.pipe)->env, env) != COR_U_SUCCEED)
        {
            // unification failed - backtrack
            iterator.pipe->stackmark0.pop_to();

            // Try the next rule
            if (!(iterator.pipe->NextRule(md))) {
                // if we have exhausted all the rules, give up
                iterator.set_no_match();
                delete iterator.pipe;
                iterator.pipe = NULL;
#                ifdef DO_TRACE
                    if (corexEnv->dbg_get_next) {
                        fprintf(corexEnv->trace_file,
                            "PipelinedDerivedRelation: get_next failed on %s(",
                            name->string());
                        iterator.arg_list.print(iterator.bindenv,
                            corexEnv->trace_file);
                        fprintf(corexEnv->trace_file, ")\n");
                    }
#                endif
                //return NULL;
                FAIL_RETURN;
            }
            tempArg = (iterator.pipe)->rInfo->arg_list;
            newArg = tempArg[(iterator.pipe)->rInfo->num_literals];
        }
    }
    // Success
    iterator.reset_no_match();
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "PipelinedDerivedRelation: get_next succeeded on %s(",
                name->string());
            iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")\n");
        }
#    endif
    
    //return NULL;
    FAIL_RETURN;
}

#undef FAIL_RETURN

/*****************************************************************************/
void corPipelinedDerivedRelation::print_name(FILE *outf)
{
    char buf[500];

    sprintf(buf, "%s", corSymbolString(name));
    fprintf(outf, "%15s/%d    : (pipelined) \n", buf, arity());
}

/*****************************************************************************/
static void initRulePointers(corTable* F, corQFModuleData* md)
{
    corModulePred *predicates = md->predicates->next;
    F->moduledata = md;
    
    for (; predicates; predicates = predicates->next) {
        Name name = predicates->pred->name;
        int arity = predicates->pred->arity;
        int offset = predicates->offset;

        switch (predicates->kind) {
        case QueryPredKind:
            F->rels[offset].AssignRel(corFind_External_Relation(name, arity));
            break;

        case LocalPredKind:
            // We create PipelinedDerivedRelations for local
            // predicates only. No tuples will be actually inserted into any of 
            // these Relations, but we still need it so as to allow us to
            // do a get_next() on it
            
            F->rels[offset].AssignRel(cormakePipelinedRelation(
                (corQFPipelinedModuleData *) md,name,arity));
            break;

        case ExternalPredKind:
            F->rels[offset].AssignRel(corFind_External_Relation(name, arity));
            
            struct corAnnoStruct *cur;
            for (cur= predicates->index_annos; cur; cur=cur->next) {
                if (cur->anno->arglist2) {
                     F->rels[offset].relPtr->add_index(cur->anno->arglist1,
                        cur->num_vars, cur->anno->arglist2);
                }
                else {
                    F->rels[offset].relPtr->add_index(cur->anno->arglist1) ;
                }
            }
            
            F->rels[offset].relPtr->check_subsum = predicates->check_subsum;
             
            break;

         case MagicQueryPredKind:
         default:
            fputs("This cannot happen.\n",corexEnv->error_file);
            break;
         }
    }
}

/*****************************************************************************/

