/*****************************************************************************/
//    BPlusTreeIndex.C
/*****************************************************************************/

#ifdef WITH_PERSISTENCE

#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"


/*****************************************************************************/
void corBPlusTreeIndex::create(corBitVector *bv, int arity, int unique)
{
    if (!SM_initialized) return;

    for (int i = 0; i < arity; i++)
        this->arity += (attrs[i] = bv->test(i));    

    int e = sm_CreateIndex(SM.volid(), SM.bufGroup(), SM_BTREENDX,
        SM_string, COR_MAX_KEY_LEN, sizeof(OID), unique, &iid);
    ErrorCheck(e, "BPlusTreeIndex::create");
}

/*****************************************************************************/
void corBPlusTreeIndex::insert(USERDESC *u_desc, corSchema &schema)
{
    if (!SM_initialized) return;

    char value_buff[COR_MAX_KEY_LEN];
    int buff_offset = 0;

    int offset = 0;
    int str_len = 0;
    int data_len = 0;
    int e;

    // For safety, check that the index is valid
    ASSERT(valid());

    // Fill value_buff with the key value.
    // This is done by copying the appropriate attributes into value_buff
    for (int i = 0; i < schema.arity; i++) {
        switch(schema[i]) {
        case COR_STRING_DBTYPE:
            memcpy((char *)&str_len, (char *)(u_desc->basePtr+offset),
                sizeof(int));
            offset = offset + sizeof(int);
            data_len = str_len * sizeof(char);
            break;

        case COR_INT_DBTYPE:
            data_len = sizeof(int);
            break;

        case COR_DOUBLE_DBTYPE:
            data_len = sizeof(double);
            break;

        default:
            CORAL_error(COR_BAD_SCHEMA, "INVALID field reached",
                "BPlusTree::insert");
            return;
            // This should not happen !!
        }

        if ((buff_offset + data_len < COR_MAX_KEY_LEN) && attrs[i]) {
            memcpy((char *)(value_buff + buff_offset),
                (char *)(u_desc->basePtr + offset), data_len );
            buff_offset += data_len;
        }
        offset += data_len;
    }

    KEY key;
    key.length = (TWO)(buff_offset);
    key.valuePtr = (void *)(value_buff);
    e = sm_InsertEntry(&iid, SM.bufGroup(), &key, &(u_desc->oid));
    ErrorCheck(e, "sm_InsertEntry");
}

/*****************************************************************************/
void corBPlusTreeIndex::delete_entry(USERDESC *u_desc, corSchema &schema)
{
    if (!SM_initialized) return;

    char value_buff[COR_MAX_KEY_LEN];
    int buff_offset = 0;

    int offset = 0;
    int str_len = 0;
    int data_len = 0;
    int e;

    // For safety, check that the index is valid
    ASSERT(valid());

    // Fill value_buff with the key value.
    // This is done by copying the appropriate attributes into value_buff
    for (int i = 0; i < schema.arity; i++) {
        switch(schema[i]) {
        case COR_STRING_DBTYPE:
            memcpy((char *)&str_len, (char *)(u_desc->basePtr+offset),
                sizeof(int));
            offset = offset + sizeof(int);
            data_len = str_len * sizeof(char);
            break;

        case COR_INT_DBTYPE:
            data_len = sizeof(int);
            break;

        case COR_DOUBLE_DBTYPE:
            data_len = sizeof(double);
            break;

        default :
            CORAL_error(COR_BAD_SCHEMA, "INVALID field reached",
                "BPlusTree::insert");
            return;
            // This should not happen !!
        }

        if ((buff_offset + data_len < COR_MAX_KEY_LEN) && attrs[i]){
            memcpy((char *)(value_buff + buff_offset),
                (char *)(u_desc->basePtr + offset), data_len );
            buff_offset += data_len;
        }
        offset += data_len;
    }

    KEY key;
    key.length = (TWO)(buff_offset);
    key.valuePtr = (void *)(value_buff);
    e = sm_RemoveEntry(&iid, SM.bufGroup(), &key, &(u_desc->oid));
    ErrorCheck(e, "sm_RemoveEntry");
}

/*****************************************************************************/
corBPlusTreeIndex::corBPlusTreeIndex()
{
    INVALIDATE_IID(iid); 
    for (int i = 0; i < COR_MAX_ARGLIST_SIZE; i++)
        attrs[i] = 0;
    arity = 0;
}

/*****************************************************************************/
int corBPlusTreeIndex::matches(char *attr_str, int arity)
{
    return !(memcmp(attrs, attr_str, arity));
}

/*****************************************************************************/
void corBPlusTreeIndex::destroy() 
{
    if (!SM_initialized) return;

    int e = sm_DestroyIndex(&iid, SM.bufGroup());
    if (ErrorCheck(e, "sm_DestroyIndex")) return;
    INVALIDATE_IID(iid); 
    for (int i = 0; i < COR_MAX_ARGLIST_SIZE; i++)
        attrs[i] = 0;
};        

/*****************************************************************************/
void corBPlusTreeIndex::corprint_adornment(FILE *outf)
{
    for (int i = 0; i < arity; i++)
        fprintf(outf, "%c", ((attrs[i]) ? 'b' : 'f'));
}

/*****************************************************************************/
int corBPlusTreeIndex::begin_load()
{
    return sm_BeginIndexLoad(&iid, SM.bufGroup(), SM.tempvolid(), 10);
}

/*****************************************************************************/
int corBPlusTreeIndex::end_load()
{
    return sm_EndIndexLoad(&iid);
}

/*****************************************************************************/
int corBPlusTreeIndex::abort_load()
{
    return sm_AbortIndexLoad(&iid);
}

/*****************************************************************************/

#endif WITH_PERSISTENCE

