// *************************************************************************
//	Adorn.C
// *************************************************************************

#include "Adorn.h"

#include "Print.h"
#include "SipGen.h"

#include "Rule.h"
#include "Globals.h"
#include "Annotations.h"
#include "RelationMisc.h"		// has_grouping_args
#include "ExportInfo.h"
#include "Sips.h"
#include "GroupingArg.h"
#include "Unparse.h"
#include "Interface.h"
#include "CArg.h"

// *************************************************************************

static int group_rule_num = 0;
static int adorn_type = 0;

// *************************************************************************

static void deepen_adorn(struct corcollection *pre_ad_rules,
	struct corcollection * ad_rules, Name name, corBitVector& adorn);
static int adorn_rule(struct corcollection *pre_ad_rules, corClause *rule,
	struct corcollection *ad_rules, Name name, corBitVector& adorn);
static void adorn_rule_preds(struct corcollection *pre_ad_rules,
	corClause *adrule, int p, corLiteral *pred, struct corcollection *ad_rules);
static int isconst2(corArg* a);
static int isbound(corCArg* a, corBitVector *u);
static int equals_pred(corLiteral *lit1, corLiteral *lit2);

// SEE NOTE!
static corLiteral *copy_pred(corLiteral *prototype);

// *************************************************************************
/*
	Construct the adorned system.  adorn_type==1: construct as described in
	"On the Power of Magic", adorn_type==2: as in "Magic Templates:
	A Spellbinding Approach to Logic Programs".

	Type 2 is different from type 1 as follows:
		- bound query args are any arg position that contains a function
			symbol, constant, or repeated variable.
		- to compute the adornment of a derived predicate, an argument is
			considered bound if *some* variable appearing in it is in the
			union of sip labels.
*/

int adorn(struct corcollection *pre_ad_rules, struct corcollection * ad_rules)
{
	corBitVector adorn;
	
	if (corparserStruct.CurModule.ExportedPreds == NULL) {
		fprintf(stderr, "'export' missing or bad format.\n");
		return 0;
	}
	else {
		for (corExportInfo *export_info = corparserStruct.CurModule.ExportedPreds;
			export_info != NULL; export_info = export_info->next)
		{
			deepen_adorn(pre_ad_rules, ad_rules, 
			export_info->pred_name, export_info->adornment);
		}
	}
 
	return 1;
}

// *************************************************************************
static void deepen_adorn(struct corcollection *pre_ad_rules,
	struct corcollection * ad_rules, Name name, corBitVector& adorn)
{
	FOR_EACH_RULE_FOR(rule, name, ad_rules) {
		if (rule->head->adorn == adorn)
			return;
	} END_EACH_RULE;

	//	Find rules that match name and adorn.
	FOR_EACH_RULE_FOR(rule, name, pre_ad_rules) {
		if (!adorn_rule(pre_ad_rules, rule, ad_rules, name, adorn))
			break;
	} END_EACH_RULE;

	//	Update the annotations
	corupdate_annotations(name, adorn);
}

// *************************************************************************
static int adorn_rule(struct corcollection *pre_ad_rules, corClause *rule,
	struct corcollection *ad_rules, Name name, corBitVector& adorn)
{
	struct corsips *s;

	//	Build the skeleton of the new adorned rule
	corClause* adrule = corcopy_rule(rule);
	adrule->head = copy_pred(rule->head);
	adrule->head->base_pred = 0;
	adrule->head->adorn = adorn;
	coralloc_preds(adrule, rule->num_preds);

	//	attach only the appropriate sip
	adrule->sips = new corsips;
	adrule->sips->adorn = adorn;
	coradd_rule(adrule, ad_rules);
	adrule->sips->next = NULL;

	//	For this rule, find its sip with the same adornment as the
	//	worklist predicate
	int sips_complete = 1;
	struct corsip *asip;
	for (s=rule->sips; s; s=s->next) {
		if (s->adorn == adorn) {
			for (int pred_num=0; pred_num < rule->num_preds; pred_num++) {
				for (asip=s->sip_list;asip;asip=asip->next) {
					if (asip->head==pred_num+1)
						break;
				}
				if (!asip)
					sips_complete = 0;
			}
			break;
		}
	}
	if (s != NULL && sips_complete)
		adrule->sips->sip_list = s->sip_list;
	else {
		if (rule->head->args.count() > COR_MAX_ARGLIST_SIZE) {
			fprintf(stderr,
				"Error: Rule (line %d) has too many args (%d) - ignored\n",
				rule->line_number, rule->head->args.count());
			fflush(stderr);
			return 1;
		}
		struct ClauseInfo rule_info[1];
		setup_vars(rule, rule_info);
		gen_sip_rule(rule_info, adorn);
		cleanup_vars(rule_info);
		for (s=rule->sips; s; s=s->next) {
			if (s->adorn == adorn)
				 break;
		}
		if (!s) {
			fprintf(stderr, "Internal Error: no sip matching adornment ");
			corprint_adornment(adorn, stderr);
			fprintf(stderr, " for %s (line %d)\n", corSymbolString(name),
				rule->line_number);
			return 0;
		}
		adrule->sips->sip_list = s->sip_list;
	}

	//	Now compute the adornment for each derived predicate in the
	//	rule's right-hand-side.
	for (int p=0; p<rule->num_preds; ++p) {
		adorn_rule_preds(pre_ad_rules, adrule, p, rule->preds[p], ad_rules) ;
	}

	//	Now handle any head deletes
	adrule->head_deletes_count = rule->head_deletes_count;
	adrule->head_deletes = new corLiteral* [adrule->head_deletes_count];
	for (int hdnum=0; hdnum < rule->head_deletes_count; hdnum++) {
		adrule->head_deletes[hdnum] = copy_pred(rule->head_deletes[hdnum]);
		adrule->head_deletes[hdnum]->base_pred = 0; 
		// So adorn will get printed

		int rulepos;
		for (rulepos=0; rulepos < rule->num_preds; rulepos++) {
			if (equals_pred(rule->preds[rulepos], rule->head_deletes[hdnum]))
				break;
		}
		if (rulepos < rule->num_preds) {
			adrule->head_deletes[hdnum]->adorn = adrule->preds[rulepos]->adorn;
		}
	}
	return 1;
}

// *************************************************************************
static void adorn_rule_preds(struct corcollection *pre_ad_rules,
	corClause *adrule, int p, corLiteral *pred, struct corcollection *ad_rules)
{
    struct corsip *arc;

	//	Add the predicate to the adorned rule
	adrule->preds[p] = copy_pred(pred);
	if (pred->base_pred) {
		adrule->preds[p]->adorn.setlen(0);
	}
	else {
		//	Compute the union of labels of the sip arcs entering the node
		//	labelled by pred
		corBitVector u;
		for (arc=adrule->sips->sip_list; arc; arc=arc->next) {
			//	sub 1 since p indexes rhs only (not head)
			if (arc->head-1 == p) {
				u.or_bits(&arc->bound_vars);
			}
		}

		//	For each arg in pred, if all vars in that arg are in the
		//	union, mark the arg as bound
		int i = 0;
		FOR_EACH_ARG(a, pred->args) {
			if (isbound(a, &u)) {
				adrule->preds[p]->adorn.set(i);
			}
			i++;
		} END_EACH_ARG

		//	Find best allowed adornment
		set_best_allowed_adorn(adrule->preds[p]);
		deepen_adorn(pre_ad_rules, ad_rules, pred->pred,
			adrule->preds[p]->adorn);
	}
}

// *************************************************************************
/*
	Recursively traverse the given argument and return 1 if at least one
	sub-arg is a constant or a function symbol.
*/

static int isconst2(corArg* a)
{
	switch (a->kindof()) {
		case COR_VARIABLE:
		case COR_DONTCARE:
			return 0;
		default:
			return 1;
	}
}

// *************************************************************************
/*
	Type 1:
		Recursively traverse the given argument and return 1 if all
		sub-args are in the set of variables represented by the BitVector u

	Type 2:
		Recursively traverse the given argument and return 1 if there is at
		least one sub-arg that is in the set of variables represented by the
		BitVector u
*/

static int isbound(corCArg* a, corBitVector *u)
{
	if ((adorn_type == 2) || (corparserStruct.CurModule.UseReturnUnify)) {
		switch (a->kindof()) {
			case COR_VARIABLE:
				return u->test(a->VarArgP()->var);
			case COR_SYMBOL_CONST:
			case COR_NUM_CONST:
				return 1;
			case COR_DONTCARE:
				return 0;
			case COR_FUNCTOR:
				FOR_EACH_ARG(arg, (a->FuncArgP())->args) {
					if (isbound(arg, u)) return 1;
				} END_EACH_ARG
				return 0;
		}
	}
	else {
		switch (a->kindof()) {
			case COR_VARIABLE:
				return u->test((a->VarArgP())->var);
			case COR_SYMBOL_CONST:
			case COR_NUM_CONST:
				return 1;
			case COR_DONTCARE:
				return 0;
			case COR_FUNCTOR:
				FOR_EACH_ARG(arg, (a->FuncArgP())->args) {
					if (!isbound(arg, u)) return 0;
				} END_EACH_ARG
				return 1;
		}
	}
	return 1;
}

// *************************************************************************
static int equals_pred(corLiteral *lit1, corLiteral *lit2)
{
	if (lit1->name() != lit2->name() || lit1->arity() != lit2->arity())
		return 0;

	for (int i=0; i < lit1->arity(); i++) {
		if (!lit1->args[i].equals(lit2->args[i]))
			return 0;
	}
	return 1;
}

// *************************************************************************
void set_best_allowed_adorn(corLiteral *lit)
{
	corPredAnnotations *cur = corparserStruct.AllowedAdornList;

	//	This method is a modified version of the method used in
	//	iterative-rel.C

	int best_shortfall = 1000;	// Infinity
	corBitVector* best_adorn = NULL;
	int restrictions = 0; // Adornments restricted for this predicate
	for (;cur != NULL; cur = cur->next()) {
		if (cur->pred == lit->predicate->name) {
			int i = lit->adorn.len();
			if (cur->adorn->len() != i) 
				continue; 
			//	An adornment restriction has been specified for this predicate
			restrictions = 1;
			int shortfall = 0;
			while (--i >= 0) {
				if (!lit->adorn.test(i) && cur->adorn->test(i))
					break; // Allowed adorn overbound
				if (lit->adorn.test(i) && !cur->adorn->test(i))
					shortfall++;
			}
			if (i >= 0) {
				//	Allowed adorn is overbound  - ignore it
				continue;
			}
			if ((shortfall < best_shortfall) &&
				(!is_disallowed_adorn(cur->pred, *(cur->adorn))))
			{
				best_adorn = cur->adorn;
			}
		}
	}
	if (!restrictions) return;
	if (!best_adorn) {
		fprintf(stderr, "Warning: adornment for %s does not match any allowed adornment.\n%s\n", lit->predicate->name->string(), "				-- adornment restrictions ignored.");
		return;
	}
	lit->adorn.init(*best_adorn);
	return;
}

// *************************************************************************
int is_disallowed_adorn(Name name, corBitVector &adorn)
{
	corPredAnnotations *cur = corparserStruct.DisallowedAdornList;
	int adorn_length = adorn.len();

	for(; cur != NULL; cur = cur->next()) {
		if (cur->pred == name) {
			if (cur->adorn->len() != adorn_length) 
				continue;
			else if (*(cur->adorn) == adorn)
				return 1;
		}
	}
	return 0;
}

// *************************************************************************
/*
	The following code is to ensure that a rule with grouping in the head
	and more than one body predicate is split into two rules as follows:
		p(tbar, <Y>) :- l1, l2, ... ln.  is split up as 
		(1): p(Xbar, <Y>) :- pprime(Xbar, Y).  and
		(2): pprime(tbar, Y) :- l1, l2, ... ln.
	This is to ensure correctness of the evaluation of the magic rewritten
	program.  It also ensures that grouping args are not bound in the
	adorned program.  Consequently, magic works for programs with grouping.
*/

void preprocess_for_grouping(struct corcollection *orig_rules,
	struct corcollection *pre_ad_rules)
{
	int j;
	char *disallowed_head_adorn, *possible_allowed_head_adorn;
	corClause *preadrule;
	Name new_pred_name;
	int new_pred_arity;

	group_rule_num = 0;
	FOR_EACH_RULE(rule, orig_rules) {
		corLiteral *head = rule->head;
		if (corhas_grouping_args(head->args)) {
			if (rule->num_preds == 0) { 
				fprintf(stderr, "ERROR: Cannot have grouping rule with ");
				fprintf(stderr, "empty body.\n");
				return;
			}
			else {
				// valid grouping rule in the program
				// FIRST emit grouping argument restrictions
				disallowed_head_adorn = init_any_adorn(head->args.count());
				possible_allowed_head_adorn=init_any_adorn(head->args.count());

				int i = 0;
				FOR_EACH_ARG(a, head->args) {
					if (a->kindof()==COR_GROUPING) {
						disallowed_head_adorn[i] = 'b';
						possible_allowed_head_adorn[i] = 'f';
					}
					i++;
				} END_EACH_ARG

				coradd_disallowed_adornment(corparserStruct, head->pred,
					disallowed_head_adorn, head->args.count());
				coradd_allowed_adornment(corparserStruct, head->pred,
					possible_allowed_head_adorn, head->args.count());
				delete disallowed_head_adorn;
				delete possible_allowed_head_adorn;
				 
				// NEXT rewrite grouping rules into two rules
				if (rule->num_preds > 1) {
					new_pred_name = build_ad_grouped_pred_name(head);
					new_pred_arity = grouped_pred_arity(head);
					create_grouping_rule_1(rule, new_pred_name, new_pred_arity,
						pre_ad_rules);
					create_grouping_rule_2(rule, new_pred_name, new_pred_arity,
						pre_ad_rules);
					continue;
				}
			}
		}

		//	Either there are no grouping args in the head or the body
		//	had exactly one predicate -- the same thing needs to be done.
		//	Just copy out the rule.
		preadrule = corcopy_rule(rule);
		preadrule->head = rule->head;
		coralloc_preds(preadrule, rule->num_preds);
		for (j=0; j <rule->num_preds; j++) {
			preadrule->preds[j] = rule->preds[j];
		}
		preadrule->head_deletes_count = rule->head_deletes_count;
		preadrule->head_deletes = rule->head_deletes;
		coradd_rule(preadrule, pre_ad_rules);
	} END_EACH_RULE
}

// *************************************************************************
Name build_ad_grouped_pred_name(corLiteral *head)
{
	char *head_name = head->pred->string();
	int head_name_length = strlen(head_name);

	char *buf = new char [head_name_length+10];
	group_rule_num++; // this gives it the unique numbering

	sprintf(buf, "grp_%d_",group_rule_num);
	char *buf1 = buf + strlen(buf);
	strcpy(buf1, head_name);
	buf1[head_name_length] = '\0';
	
	return corEnterSymbol(buf);
}

// *************************************************************************
int grouped_pred_arity(corLiteral *head)
{
	int ret_val = 0;
	FOR_EACH_ARG(a, head->args) {
		if (a->kindof() != COR_GROUPING) {
			ret_val++;
		}
		else {
			// it is a grouping argument
			ret_val += a->GroupingP()->arity();
		}
	} END_EACH_ARG
	return ret_val;
}

// *************************************************************************
char *init_any_adorn(int num)
{
    char *ret_val = new char[num+1];

    for (int i = 0; i < num; i++) 
		ret_val[i] = 'a';
    ret_val[num] = '\0';
    return ret_val;
}

// *************************************************************************
/*
	This adds the rule of the form
		p(Xbar, <Y>) :- pprime(Xbar, Y)
	obtained from 
		p(tbar, <Y>) :- l1, l2, ... ln
*/

void create_grouping_rule_1(corClause *rule, Name new_pred_name,
	int new_pred_arity, struct corcollection *pre_ad_rules)
{
	corClause *grouping_rule_1 = corcopy_rule(rule);
	corLiteral *head = rule->head;
	int j, k;
	corCArg *arg_ptr, *grp_arg_ptr;
	corGrouping *new_groupingarg;

	// create the new head literal
	corLiteral *new_head = new corLiteral(corparserStruct, head->pred,
		head->args.count());
	j = 0;
	arg_ptr = & (new_head->args[0]);
	FOR_EACH_ARG(a, head->args) {
		if (a->kindof() != COR_GROUPING) {
			*arg_ptr++ = make_var(j++);
		}
		else {
			// it was a grouping argument earlier and it has to be now
			new_groupingarg = corGrouping::New(a->GroupingP()->arity());
			new_groupingarg->kind =  a->GroupingP()->kind;

			grp_arg_ptr = &((new_groupingarg->nested_args())[0]);

			for (k=0; k < new_groupingarg->arity(); k++) {
				*grp_arg_ptr++ = make_var(j++);		
			}
			*arg_ptr++ = new_groupingarg;
		}
	} END_EACH_ARG

	// allocate the head
	grouping_rule_1->head = new_head;

	// create the new body literal
	corLiteral *new_body = new corLiteral(corparserStruct, new_pred_name,
		new_pred_arity);
	arg_ptr = & (new_body->args[0]);
	for (j=0; j < new_pred_arity; j++) {
		*arg_ptr++ = make_var(j);
	}

	// allocate the body
	coralloc_preds(grouping_rule_1, 1);
	grouping_rule_1->preds[0] = new_body; 

	coradd_rule(grouping_rule_1, pre_ad_rules);
}

// *************************************************************************
/*
	This implicitly adds the rule of the form
		pprime(tbar, Y) :- l1, l2, ... ln.
	obtained from 
		p(tbar, <Y>) :- l1, l2, ... ln
*/

void create_grouping_rule_2(corClause *rule, Name new_pred_name,
	int new_pred_arity, struct corcollection *pre_ad_rules)
{
	int j;
	corClause *grouping_rule_2;
	corCArg *arg_ptr, *grp_arg_ptr;
	corLiteral *head = rule->head;

	grouping_rule_2 = corcopy_rule(rule);

	// create the new head 
	corLiteral *new_head = new corLiteral(corparserStruct, 
									 new_pred_name, new_pred_arity);
	arg_ptr = &(new_head->args[0]);
	FOR_EACH_ARG(a, head->args) {
		if (a->kindof() != COR_GROUPING) {
			*arg_ptr++ = *a;
		}
		else { // it is a grouping argument 
			grp_arg_ptr = &((a->GroupingP()->nested_args())[0]);
			for (j=0; j < a->GroupingP()->arity(); j++) {
				*arg_ptr++ = *grp_arg_ptr++;
			}
		}
	} END_EACH_ARG

	// allocate the head
	grouping_rule_2->head = new_head;

	// copy out the body
	coralloc_preds(grouping_rule_2, rule->num_preds);
	for (j=0; j <rule->num_preds; j++) {
		grouping_rule_2->preds[j] = rule->preds[j];
	}
	coradd_rule(grouping_rule_2, pre_ad_rules);
}

// *************************************************************************
//	NOTE: THIS NEEDS TO BE FIXED
//	ADD A MEANS OF COPYING A LITERAL TO LITERAL.Ch in CORAL

static corLiteral *copy_pred(corLiteral *prototype)
{
	corLiteral *pred = new corLiteral(prototype->args.count());
	*pred = *prototype;
    pred->adorn.init(prototype->adorn); // Fake copy constructor.
    return pred;
}

// *************************************************************************

