// *************************************************************************
//    Main.C
//    Implements Magic, Supplementary Magic and Context Rule Rewriting
//    (as in Kemp, Ramamohanrao & Somogyi(VLDB 90))
// *************************************************************************

#include "Defunct.h"
#include "Linearity.h"
#include "Adorn.h"
#include "Print.h"
#include "SipOrder.h"
#include "SupMagic.h"
#include "FullMagic.h"
#include "RuleFactor.h"

#include "CoralInit.h"
#include "Collection.h"
#include "ModuleInfo.h"
#include "Scc.h"
#include "Unparse.h"
#include "RelationMisc.h"
#include "Compile.h"
#include "ParseUtils.h"
#include "BasePreds.h"
#include "Globals.h"
#include "ExportInfo.h"
#include "Annotations.h"
#include <assert.h>

// *************************************************************************
//
//  Notes:
//  1) What do I do about opt_share_functors and debugmagic ??
//  2) Factoring and SupMagicIndexing do not co-exist. If both options are set,
//      supplementary magic with supmagicindexing is done. Factoring is not.
//
// *************************************************************************

extern void preprocess_for_grouping(struct corcollection *orig_rules, 
				    struct corcollection *pre_ad_rules);
/*				    
extern void add_make_index_annotation(PredAnnotations *&, Symbol *pred, 
				      Symbol *adorn_string,
				      ArgList *al1, ArgList *al2);
*/

// *************************************************************************

FILE *inf  = stdin;
FILE *outf = stdout;

int    optAdorn=0, optFactor=0, optDefunct=0;

// Used as extern by funct/defunct code
int    orig_rule_count = 0;

struct corcollection pre_ad_rules[1];
struct corcollection factor_rules[1];
struct corcollection magic_rules[1];

// *************************************************************************

extern int show_prompt;
extern corScc *corSccList;

// *************************************************************************

extern "C" int yyparse();

// *************************************************************************

static void printUsage(char *arg0);
static void mark_scc(corScc *scc, LinearKind kind);
static void check_ord_search(corParserStruct &corparserStruct);
static void find_status();
static void generate_index_annos(struct corcollection *rules,
    corPredAnnotations*& index_annos);
static void clear_sips(struct corcollection* rules);

void do_factor_rewriting();
void add_magic_rule_labels (struct corcollection *factor_rules);

// *************************************************************************
main(int ac, char** av)
{
  int i;
  char *infile = "stdin", *outfile = NULL;

  init_coral(av[0]);

  for (i=1; i<ac && (*av[i]=='-'); ++i) {
      if (!strcmp(av[i]+1, "a")) optAdorn = 1;
      else if (!strcmp(av[i]+1, "d")) optDefunct = 1;
      else if (!strcmp(av[i]+1, "f")) optFactor = 1;
      else if (!strcmp(av[i]+1, "i"))
          infile = av[++i];
      else if (!strcmp(av[i]+1, "o"))
          outfile = av[++i];
      else if (!strcmp(av[i]+1, "factor")) {
          if (++i < ac) corexEnv->C_use_factoring_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "magic")) {
          if (++i < ac) corexEnv->C_use_magic_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "supmagic")) {
          if (++i < ac) corexEnv->C_use_supplementary_magic_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "pipelining")) {
          if (++i < ac) corexEnv->C_use_pipelining_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "supmagic_indexing")) {
          if (++i < ac) corexEnv->C_sup_magic_indexing_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "ordsearch")) {
          if (++i < ac) corexEnv->C_use_ordsearch_default = atoi(av[i]);
      }
      else if (!strcmp(av[i]+1, "eqo")) {
          if (++i < ac) corexEnv->C_use_exist_opt_default = atoi(av[i]);
      }
      else {
          printUsage(av[0]);
          exit(1);
      }
  }

  if (strcmp(infile,"stdin") != 0)
      if ((inf = fopen(infile,"r")) == NULL){
         fprintf(stderr,"Could not open file %s for input.\n",infile);
         exit(1);
      }
  corparseStack->handle_new_file(inf, infile);
    

  if (outfile){
      if ((outf = fopen(outfile,"w")) == NULL){
         fprintf(stderr,"Could not open file %s for output. (using stdout)\n",
                outfile);
         outf = stdout;
      }
  }

  corexEnv->C_print_prompt_default = 0;
  corexEnv->C_exec_mode = COR_DOING_FACTOR;

  for(;;) {
      corparserStruct.rule_list = NULL;
      //exEnv->C_at_end_module = 0;

      if (yyparse()) {
          fputs("syntax error -- quitting\n",stderr);
          break;
      }

      if (corexEnv->C_at_end_module && corparserStruct.rule_list)
          do_factor_rewriting();
      else if (corparserStruct.rule_list) {
          fputs("CORAL::factor: Warning - end_module missing\n",stderr);
          do_factor_rewriting();
      }
      else break;
  }
  exit_coral();
  exit(0);
}

// *************************************************************************
static void printUsage(char *arg0)
{
   fprintf(stderr,
        "\nUsage: %s [-l] [-a] [-c] [-r] [-i infile] [-o outfile]\n\n",arg0);
   fputs("    -a    Print the program with initial adornments.\n",stderr);
   fputs("    -d    Print the Extended Datalog program.\n",stderr);
   fputs("    -f    Print the program with factoring status of rules.\n",
            stderr);
   fputs("    -i    Use infile instead of stdin.\n",stderr);
   fputs("    -o    Use outfile instead of stdout.\n\n",stderr);
}

// *************************************************************************
static void mark_scc(corScc *scc, LinearKind kind)
{
    for (corPredicate *pred = scc->preds; pred != NULL; pred = pred->next) {
        corClause *clause = pred->clauses;
        for (; clause != *pred->clause_tail; clause = clause->pnext) {
            if (!clause->seen) continue;
            clause->factor_status = (int)kind;
        }
    }
}

// *************************************************************************
/*
    This is used to determine whether the module needs to be computed using
    Ordered Search.  There are two cases:
        (1) If there is aggregation in the head, and a derived literal in the
            body.
        (2) If there is genuine negation in the body of the rule, this should
            work for both magic and supmagic.  I'm not sure about supmagic
            -- Praveen
*/
// *************************************************************************

static void check_ord_search(corParserStruct &corparserStruct)
{
    if (corparserStruct.CurModule.UseOrdSearch) return;
    
    FOR_EACH_RULE(adrule, corparserStruct.rules) {
        for (int i=0; i < adrule->num_preds; ++i) {
            if (!adrule->preds[i]->base_pred) {
                // check if there is aggregation in the head
                if (corhas_grouping_args(adrule->head->args)) {
                    COR_set_use_ordsearch(corparserStruct.CurModule, 1);
                    return;
                }
                // check if there is genuine negation in the head of the rule
                if (adrule->preds[i]->negated) {
                    // negated and derived predicates
                    COR_set_use_ordsearch(corparserStruct.CurModule, 1);
                    return;
                }
            }
        }
    } END_EACH_RULE
}


// *************************************************************************
/*
    Assigns linear status to the rules that define the predicates reachable
    from the exported predicates. For each exported predicate, an SCC analysis 
    is performed, and the rules in each scc are analysed independently.
    Factoring can be performed at any level of such stratified rules.
    For those sccs that cannot be factored, magic transformation is done.

    PRAVEEN: What happens if there is more than one such exported pred?
    Answer: There is a fixed scc structure for the rules, irrespective of
    the exported predicates. So the same rule cannot be given one linear
    status for one exported predicate, and a different status for another.
*/
// *************************************************************************

static void find_status()
{

    for (corExportInfo *export_info = corparserStruct.CurModule.ExportedPreds;
        export_info != NULL; export_info = export_info->next)
    {
        Name query_symbol;
        Name exported_name = export_info->pred_name;
        int query_arity;

        if (export_info->adorn_symbol == 0) {
            query_arity = 0;
            query_symbol = exported_name;
        }
        else {
            query_arity = (int) export_info->adorn_symbol->length();
            query_symbol = corEnterSymbol(exported_name->string());
        }

        export_info->predicate = corLookupPredicate(corparserStruct, query_symbol,
            query_arity);
        if (!export_info->predicate) {
            fprintf(stderr,
                "Error: Exported predicate %s not defined in module\n",
                query_symbol->string());
            continue;
        }

        //    Do Scc (Strongly Connected Components) analysis.
        FOR_ALL_RULES(clause) {
            clause->seen = 0;

            clause->recursive = 0;
            for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                clause->preds[ipred]->predicate->work = NULL;
                clause->preds[ipred]->predicate->mark = 0;
            }
        } END_EACH_RULE
        corinitialize_scc_structs();

        corfind_predicates_invoked_from(export_info->predicate);
        cornumber_sccs(corSccList); // added by Ken

        if (corparserStruct.CurModule.NestedSccEval)
          continue;

        for (corScc* scc = corSccList; scc != NULL; scc = scc->next) {
            corPredicate *pred; corClause *clause; int ipred;

            //    Use the scc->recursive field to specify whether the scc
            //    contains more than one recursive predicate.

            //    If the module is not to be factored, do magic instead.

            if ((!corparserStruct.CurModule.UseFactoring) ||
                (corparserStruct.CurModule.SupMagicIndexing))
            {
                mark_scc(scc, COR_NON_LINEAR);
                continue;
            }

            scc->recursive = 0;

            for (pred = scc->preds; pred != NULL; pred = pred->next) {
                clause = pred->clauses;
                for (; clause != *pred->clause_tail; clause = clause->pnext) {
                    if (!clause->seen) continue;
                    for (ipred = 1; ipred <= clause->num_preds; ipred++) {
                        corPredicate *child =
                            clause->preds[ipred-1]->predicate;
                        if (child->scc == scc) clause->recursive = 1;
                        if ((child->scc == scc) && (child != pred)
                            && (!clause->preds[ipred-1]->base_pred))
                        {
                            // Mutually recursive scc !!
                            scc->recursive = 1;
                            break;
                        }
                        if (scc->recursive) break;
                    }
                    if (scc->recursive) break;
                }
                if (scc->recursive) break;
              }

            if (scc->recursive) {
                mark_scc(scc, COR_NON_LINEAR);
                continue;
            }

            for (pred = scc->preds; pred != NULL; pred = pred->next) {
                if (pred->scc == scc) break;
            }

            //    At this stage, pred contains the only derived predicate
            //    of the scc.
            if (pred == NULL) {
                fputs("SOMETHING WRONG HERE !\n",stderr);
                break;
            }

            clause = pred->clauses;
            for (; clause != *pred->clause_tail; clause = clause->pnext) {
                if (!clause->seen) continue;
                if (!assign_linear_status(clause)) {
                      //    Use recursive field to signify that this scc
                      //    is not factorable.
                      scc->recursive = 1;
                      break;
                }
                else if (clause->factor_status == PSEUDO_LT_LINEAR) {
                    // As pointed out by David Kemp, pseudo-lt-linear rules
                    // are not compatible with other kinds of linear rules.
                    // Also, in the case of pseudo-lt-linear rules and
                    // lt-linear rules, context rewriting is only as
                    // efficient as magic, so might as well do magic
                    // on it.
                      scc->recursive = 1;
                    break;
                }
            }
            if (scc->recursive) mark_scc(scc, COR_NON_LINEAR);
        }
    }
}

// *************************************************************************
static void generate_index_annos(struct corcollection *rules,
    corPredAnnotations*& index_annos)
{
    char buf[COR_MAX_ARGLIST_SIZE+1];
    int i, p;
    corArgList* al1;
    corArg* tmp_arg;

    FOR_EACH_RULE(rule, rules) {
        for (p=0; p < rule->num_preds; ++p) {
            if ((rule->preds[p]->base_pred) && (rule->preds[p]->adorn.length)) {
                for (i = 0; i < rule->preds[p]->adorn.length; ++i) {
                    buf[i] = rule->preds[p]->adorn[i] ? 'b' : 'f';
                }
                buf[i] = '\0';
                tmp_arg = corEnterSymbol(buf);
                al1 = new corArgList(1);
                (*al1)[0] = tmp_arg;
                coradd_make_index_annotation(index_annos,
                    rule->preds[p]->name(), (corSymbol*)NULL,
                    al1, NULL);
            }
        }
    } END_EACH_RULE;
}


// *************************************************************************
static void clear_sips(struct corcollection* rules)
{
    FOR_EACH_RULE(rule, rules) {
        rule->sips = NULL;
    } END_EACH_RULE;
}

// *************************************************************************
void do_factor_rewriting() 
{
    TermList **terms; // For factoring

    if (corparserStruct.CurModule.UsePipelining) {
        //    If the module uses pipelining, there is really no need to do any
        //    rewriting.  However, in order to correctly generate indices, there
        //    must be an analysis of how bindings are propogated.  One way to
        //    do this is to first mark all preds as derived preds and perform
        //    the adornment phase.  This will cause the base preds to be
        //    adorned in all the possible ways.  The actual base preds can now
        //    be marked, and index annotations generated for just these preds.
        //    Finally, these index annotations are dumped along with the
        //    original module.

        corparserStruct.CurModule.PreProcessing = 0;
        mark_all_as_derived(corparserStruct.rules);
        corinit_collection(pre_ad_rules, 1);
        if (!adorn(corparserStruct.rules, pre_ad_rules)) {
            fputs("Unable to proceed with magic/factor rewriting!\n",stderr);
            exit(1);
        }

        if (optAdorn) {
            fputs("Printing adorned rules\n",stderr);
            print_factor(pre_ad_rules, stderr, 0);
            fputs("\n\n",stderr);
        }
        
        mark_basepreds(pre_ad_rules);
        generate_index_annos(pre_ad_rules, corparserStruct.make_index_annotations);
        clear_sips(corparserStruct.rules);
        print_factor(corparserStruct.rules, outf, 0);
        return;
    }

    if (!corparserStruct.CurModule.Rewriting ||
        (!corparserStruct.CurModule.UseMagic &&
            !corparserStruct.CurModule.UseSupplementaryMagic))
    {
        fprintf(stderr,
            "Warning: Magic rewriting not performed for Module %s.\n",
            corparserStruct.CurModule.name->string());
        //parserStruct.CurModule.Rewriting = 0;
        corparserStruct.CurModule.PreProcessing = 0;
        print_factor(corparserStruct.rules, outf, 0);
        return;
    }

    mark_basepreds(corparserStruct.rules);

    corinit_collection(pre_ad_rules, 1);

    //    For monotonic programs, there is no need to do ordered search.
    //    Also, if ordered search is already set, don't bother checking again.
    if (!corparserStruct.CurModule.UseOrdSearch &&
        !corparserStruct.CurModule.Monotonic)
            check_ord_search(corparserStruct);

    if (corparserStruct.CurModule.UseFactoring &&
        corparserStruct.CurModule.UseOrdSearch)
    {
        fprintf(corexEnv->error_file, "Ordered search is turned on!\n");
        fprintf(corexEnv->error_file, "Have to turn factoring off : sorry!\n");
        COR_set_use_factoring(corparserStruct.CurModule, 0);     
    }

    preprocess_for_grouping(corparserStruct.rules, pre_ad_rules);

    if (optAdorn) {
        fprintf(stderr, "Printing pre-adorned rules\n");
        print_factor(pre_ad_rules, stderr, 0);
        fprintf(stderr, "\n\n");
    }

    corinit_collection(factor_rules, 1);
    if (!adorn(pre_ad_rules, factor_rules)) {
        fprintf(stderr, "Unable to proceed with magic/factor rewriting!\n");
        exit(1);
   }

    if (optAdorn) {
        fprintf(stderr, "Printing adorned rules\n");
        print_factor(factor_rules, stderr, 0);
        fprintf(stderr, "\n\n");
    }

    corinit_collection(magic_rules, 0);

    FOR_EACH_RULE(r, factor_rules) {
        r->head->predicate->clauses = NULL;
        r->head->predicate->clause_tail = &(r->head->predicate->clauses);
    } END_EACH_RULE

    FOR_EACH_RULE(r, factor_rules) {
        corAddClauseToPredicate(r->head->predicate, r);
    } END_EACH_RULE

    if (corparserStruct.CurModule.UseFactoring) {
        //    NOTE:  It is important that the TermList be created only at
        //    this point, where the size of the adorned rule collection is
        //    known.
        orig_rule_count = factor_rules->count;
        terms = new TermList*[orig_rule_count];

        defunctify((corClause *)(factor_rules->chain), terms); 
        if (optDefunct) {
            fprintf(stderr, "Printing rules in Extended Datalog form\n");
            print_factor(factor_rules, stderr, 0);
            fprintf(stderr, "\n\n");
        }

        //    Do an SCC analysis, and assign linear status to each rule.
        find_status();

        if (corparserStruct.CurModule.UseFactoring) {
            functify((corClause *)(factor_rules->chain), terms); 
        }

        if (optFactor) {
            fprintf(stderr, "Printing rules after factoring\n");
            print_factor(factor_rules, stderr, 1);
            fprintf(stderr, "\n\n");
        }
        factor_rewrite(factor_rules, magic_rules);
    }
    else {
        siporder(factor_rules);

        if (corparserStruct.CurModule.NestedSccEval) { //Added by Ken
          // Do Scc analysis
          find_status(); 
          // classify rules
          FOR_EACH_RULE(rule, factor_rules) {
            int sup = 0;
            for (int i=0; i < rule->num_preds; ++i) {
              if (!rule->preds[i]->base_pred &&
              rule->preds[i]->predicate->scc->scc_no != 
                rule->head->predicate->scc->scc_no) 
            sup++;
            }
//            printf("Sup? %d\n", sup); fflush(stdout);
            if (sup)
              rule->supp = 1; // should be rewritten using sup_magic
            else
              rule->supp = 0; // should be rewritten using magic
            
          } END_EACH_RULE
          // perform rewriting for nested-scc evaluation
          supmagic(factor_rules, magic_rules);
          fullmagic(factor_rules, magic_rules);
          
        }
        else if (corparserStruct.CurModule.UseSupplementaryMagic) {
            supmagic(factor_rules, magic_rules);
            // supopt();
        }
        else if ( corparserStruct.CurModule.UseMagic)
            fullmagic(factor_rules, magic_rules);
    }
    //    Set @no_preprocessing to show that factoring/magic has been performed
    //parserStruct.CurModule.Rewriting = 0;
    corparserStruct.CurModule.PreProcessing = 0;

    add_magic_rule_labels(magic_rules);  // added by Ken

    print_factor(magic_rules, outf, 0);
}

// *************************************************************************
void add_magic_rule_labels (struct corcollection *factor_rules)
{
    /* find the number of original rules */
    int num_org_rules = 0;
    int tot_rules = 1;
    int tmp_rule_number;

    FOR_EACH_RULE(factorrule, factor_rules) {
    //  corprint_rule(factorrule, stdout); fflush(stdout);  // DEBUG
      if (sscanf(factorrule->clause_names->name->string(), "%d", 
             &tmp_rule_number) == 1)
          if (tmp_rule_number > tot_rules)
              tot_rules = tmp_rule_number;

      if (!(factorrule->head->kind == COR_MAGIC || 
                factorrule->head->kind == COR_SUPMAGIC))
        num_org_rules++;

    } END_EACH_RULE

    int *magic_rule_label_counter = 
             (int*) malloc((tot_rules+1) * sizeof(int));

    for (int i=1; i<=tot_rules; i++)
      magic_rule_label_counter[i] = 1;
    
    FOR_EACH_RULE(factorrule, factor_rules) {
      int org_rule_number;
      char label_str[5];
      char org_label_str[5];
      strcpy(org_label_str,"0");
      struct corClauseName *org_cname = new corClauseName();
      org_cname->name = corEnterSymbol(org_label_str);
      org_cname->negated = 0;
      org_cname->next = NULL;

      if (!(factorrule->head->kind == COR_MAGIC || 
                    factorrule->head->kind == COR_SUPMAGIC)){
        // original rule
        // Assert shows location of problem clearly
            assert(factorrule->clause_names);
            factorrule->clause_names->next = org_cname;

      } else {
        // derived rule
        sscanf(factorrule->clause_names->name->string(), "%d", 
               &org_rule_number);

        sprintf(label_str, "%d", 
            magic_rule_label_counter[org_rule_number]); 
        struct corClauseName *cname1 = new corClauseName();
        cname1->name = corEnterSymbol(factorrule->clause_names->name->string());
        cname1->negated = 0;
        struct corClauseName *cname2 = new corClauseName();
        cname2->name = corEnterSymbol(label_str);
        cname2->negated = 0;
        
        cname1->next = cname2;
        cname2->next = NULL;
        
        factorrule->clause_names = cname1;
        magic_rule_label_counter[org_rule_number]++;
      }
    } END_EACH_RULE
 
    free(magic_rule_label_counter);
}

// *************************************************************************
